# -*- coding: utf-8 -*-


class SceptreException(Exception):
    """
    Base class for all Sceptre errors
    """

    pass


class ProjectAlreadyExistsError(SceptreException):
    """
    Error raised when Sceptre project already exists.
    """

    pass


class InvalidSceptreDirectoryError(SceptreException):
    """
    Error raised if a sceptre directory is invalid.
    """

    pass


class UnsupportedTemplateFileTypeError(SceptreException):
    """
    Error raised if an unsupported template file type is used.
    """

    pass


class TemplateSceptreHandlerError(SceptreException):
    """
    Error raised if sceptre_handler() is not defined correctly in the template.
    """

    pass


class DependencyDoesNotExistError(SceptreException):
    """
    Error raised when a dependency cannot be found
    """

    pass


class DependencyStackNotLaunchedError(SceptreException):
    """
    Error raised when a dependency stack has not been launched
    """

    pass


class DependencyStackMissingOutputError(SceptreException):
    """
    Error raised if a dependency stack does not have the correct outputs.
    """

    pass


class CircularDependenciesError(SceptreException):
    """
    Error raised if there are circular dependencies
    """

    pass


class UnknownStackStatusError(SceptreException):
    """
    Error raised if an unknown stack status is received.
    """

    pass


class RetryLimitExceededError(SceptreException):
    """
    Error raised if the request limit is exceeded.
    """

    pass


class UnknownHookTypeError(SceptreException):
    """
    Error raised if an unrecognised hook type is received.
    """


class VersionIncompatibleError(SceptreException):
    """
    Error raised if configuration incompatible with running version.
    """

    pass


class ProtectedStackError(SceptreException):
    """
    Error raised upon execution of an action under active protection
    """

    pass


class UnknownStackChangeSetStatusError(SceptreException):
    """
    Error raised if an unknown stack change set status is received.
    """

    pass


class InvalidHookArgumentTypeError(SceptreException):
    """
    Error raised if a hook's argument type is invalid.
    """

    pass


class InvalidHookArgumentSyntaxError(SceptreException):
    """
    Error raised if a hook's argument syntax is invalid.
    """

    pass


class InvalidHookArgumentValueError(SceptreException):
    """
    Error raised if a hook's argument value is invalid.
    """

    pass


class CannotUpdateFailedStackError(SceptreException):
    """
    Error raised when a failed stack is updated.
    """

    pass


class StackDoesNotExistError(SceptreException):
    """
    Error raised when a stack does not exist.
    """

    pass


class ConfigFileNotFoundError(SceptreException):
    """
    Error raised when a config file does not exist.
    """

    pass


class InvalidConfigFileError(SceptreException):
    """
    Error raised when a config file lacks mandatory keys.
    """

    pass


class PathConversionError(SceptreException):
    """
    Error raised when a path is unable to be converted.
    """

    pass


class InvalidAWSCredentialsError(SceptreException):
    """
    Error raised when AWS credentials are invalid.
    """

    pass


class TemplateHandlerNotFoundError(SceptreException):
    """
    Error raised when a Template Handler of a certain type is not found
    """

    pass


class TemplateHandlerArgumentsInvalidError(SceptreException):
    """
    Error raised when the arguments passed to a Template Handler do not
    adhere to the specified JSON schema.
    """


class TemplateNotFoundError(SceptreException):
    """
    Error raised when a Template file is not found
    """

    pass


class CannotPruneStackError(SceptreException):
    """
    Error raised when an obsolete stack cannot be pruned because another stack depends on it that is
    not itself obsolete.
    """


class InvalidResolverArgumentError(SceptreException):
    """
    Indicates a resolver argument is invalid in some way.
    """



import difflib
import logging
from abc import abstractmethod
from typing import (
    NamedTuple,
    Dict,
    List,
    Optional,
    Callable,
    Tuple,
    Generic,
    TypeVar,
    Union,
)

import cfn_flip
import deepdiff
import yaml
from cfn_tools import ODict
from yaml import Dumper

from sceptre.exceptions import SceptreException
from sceptre.plan.actions import StackActions
from sceptre.stack import Stack

from botocore.exceptions import ClientError

DiffType = TypeVar("DiffType")

logger = logging.getLogger(__name__)


class StackConfiguration(NamedTuple):
    """A data container to represent the comparable parts of a Stack."""

    stack_name: str
    parameters: Dict[str, Union[str, List[str]]]
    stack_tags: Dict[str, str]
    notifications: List[str]
    cloudformation_service_role: Optional[str]


class StackDiff(NamedTuple):
    """A data container to represent the full difference between a deployed stack and the stack as
    it exists locally within Sceptre.
    """

    stack_name: str
    template_diff: DiffType
    config_diff: DiffType
    is_deployed: bool
    generated_config: StackConfiguration
    generated_template: str


def repr_str(dumper: Dumper, data: str) -> str:
    """A YAML Representer that handles strings, breaking multi-line strings into something a lot
    more readable in the yaml output. This is useful for representing long, multiline strings in
    templates or in stack parameters.

    :param dumper: The Dumper that is being used to serialize this object
    :param data: The string to serialize
    :return: The represented string
    """
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_str(data)


def repr_odict(dumper: Dumper, data: ODict) -> str:
    """A YAML Representer for cfn-flip's ODict objects.

    ODicts are a variation on OrderedDicts for that library. Since the Diff command makes extensive
    use of ODicts, they can end up in diff output and the PyYaml library doesn't otherwise calls the
    dicts like !!ODict, which looks weird. We can just treat them like normal dicts when we serialize
    them, though.

    :param dumper: The Dumper that is being used to serialize this object
    :param data: The ODict object to serialize
    :return: The serialized ODict
    """
    return dumper.represent_dict(data)


yaml.add_representer(str, repr_str)
yaml.add_representer(ODict, repr_odict)


class StackDiffer(Generic[DiffType]):
    """A utility for producing a StackDiff that indicates the full difference between a given stack
    as it is currently DEPLOYED on CloudFormation and the stack as it exists in the local Sceptre
    configurations.

    This utility compares both the stack configuration (specifically those attributes that CAN be
    compared) as well as the stack template.

    As an abstract base class, the two comparison methods need to be implemented so that the
    StackDiff can be generated.
    """

    STACK_STATUSES_INDICATING_NOT_DEPLOYED = [
        "CREATE_FAILED",
        "ROLLBACK_COMPLETE",
        "DELETE_COMPLETE",
    ]

    NO_ECHO_REPLACEMENT = "****"

    def __init__(self, show_no_echo=False):
        """Initializes the StackDiffer.

        :param show_no_echo: If True, local parameters passed that the template says are NoEcho
            parameters will be displayed; Otherwise, they will be masked in the diff.
        """
        self.show_no_echo = show_no_echo

    def diff(self, stack_actions: StackActions) -> StackDiff:
        """Produces a StackDiff between the currently deployed stack (if it exists) and the stack
        as it exists locally in Sceptre.

        :param stack_actions: The StackActions object to use for generating and fetching templates
            as well as providing other details about the stack.
        :return: The StackDiff that expresses the difference.
        """
        generated_config = self._create_generated_config(stack_actions.stack)
        deployed_config = self._create_deployed_stack_config(stack_actions)
        is_stack_deployed = bool(deployed_config)

        generated_template = self._generate_template(stack_actions)
        deployed_template = self._get_deployed_template(
            stack_actions, is_stack_deployed
        )

        self._handle_special_parameter_situations(
            stack_actions, generated_config, deployed_config
        )

        template_diff = self.compare_templates(deployed_template, generated_template)
        config_diff = self.compare_stack_configurations(
            deployed_config, generated_config
        )

        return StackDiff(
            stack_actions.stack.external_name,
            template_diff,
            config_diff,
            is_stack_deployed,
            generated_config,
            generated_template,
        )

    def _create_generated_config(self, stack: Stack) -> StackConfiguration:
        parameters = self._extract_parameters_from_generated_stack(stack)
        stack_configuration = StackConfiguration(
            stack_name=stack.external_name,
            parameters=parameters,
            stack_tags=stack.tags,
            notifications=stack.notifications,
            cloudformation_service_role=stack.cloudformation_service_role,
        )

        return stack_configuration

    def _extract_parameters_from_generated_stack(self, stack: Stack) -> dict:
        """Extracts a usable dict of parameters from the stack, performing some minor transformations
        to match the what CloudFormation does on its end.

        :param stack: The stack to extract the parameters from
        :return: A dictionary of stack parameters to be compared.
        """
        formatted_parameters = {}
        for key, value in stack.parameters.items():
            # When boto3 receives "None" for a cloudformation parameter, it treats it as if the
            # value is not passed at all. To be consistent in our diffing, we need to skip Nones
            # altogether.
            if value is None:
                continue

            try:
                if isinstance(value, list):
                    value = ",".join(item.rstrip("\n") for item in value)
                formatted_parameters[key] = value.rstrip("\n")
            # Other unexpected data can get through and this would blow up the differ
            # and lead to quite confusing exceptions being raised. This check here could
            # be removed in a future version of Sceptre if the reader class did sanity checking.
            except AttributeError:
                raise SceptreException(
                    f"Parameter '{key}' whose value is {value} "
                    f"is of type {type(value)} and not expected here"
                )

        return formatted_parameters

    def _create_deployed_stack_config(
        self, stack_actions: StackActions
    ) -> Optional[StackConfiguration]:
        try:
            description = stack_actions.describe()
        except ClientError as err:
            # This means the stack has not been deployed yet
            if err.response["Error"]["Message"].endswith("does not exist"):
                return None

        stacks = description["Stacks"]
        for stack in stacks:
            if stack["StackStatus"] in self.STACK_STATUSES_INDICATING_NOT_DEPLOYED:
                return None
            return StackConfiguration(
                parameters={
                    param["ParameterKey"]: param["ParameterValue"]
                    for param in stack.get("Parameters", [])
                },
                stack_tags={tag["Key"]: tag["Value"] for tag in stack["Tags"]},
                stack_name=stack["StackName"],
                notifications=stack["NotificationARNs"],
                cloudformation_service_role=stack.get("RoleARN"),
            )

    def _handle_special_parameter_situations(
        self,
        stack_actions: StackActions,
        generated_config: StackConfiguration,
        deployed_config: StackConfiguration,
    ):
        deployed_template_summary = stack_actions.fetch_remote_template_summary()
        generated_template_summary = stack_actions.fetch_local_template_summary()

        if deployed_config is not None:
            # Trailing linebreaks sometimes get removed by CloudFormation in certain circumstances
            # and can sometimes come from using the !file_contents resolver, but ultimately they
            # shouldn't affect the diff. We'll ignore all trailing linebreaks.
            self._remove_terminating_linebreaks_from_deployed_parameters(
                deployed_template_summary, deployed_config
            )

            # If the parameter is not passed by Sceptre and the value on the deployed parameter is
            # the default value, we'll actually remove it from the deployed parameters list so it
            # doesn't show up as a false positive.
            self._remove_deployed_default_parameters_that_arent_passed(
                deployed_template_summary, generated_config, deployed_config
            )

        if not self.show_no_echo:
            # We don't actually want to show parameters Sceptre is passing that the local template
            # marks as NoEcho parameters (unless show_no_echo is set to true). Therefore those
            # parameter values will be masked.
            self._mask_no_echo_parameters(generated_template_summary, generated_config)

    def _remove_terminating_linebreaks_from_deployed_parameters(
        self, template_summary: Optional[dict], deployed_config: StackConfiguration
    ):
        if template_summary is None:
            return

        parameter_types = {
            parameter["ParameterKey"]: parameter["ParameterType"]
            for parameter in template_summary["Parameters"]
        }

        for key, value in deployed_config.parameters.items():
            parameter_type = parameter_types[key]
            if parameter_type == "CommaDelimitedList":
                # If it's a list of strings, remove trailing linebreaks for each item
                value = ",".join([item.rstrip("\n") for item in value.split(",")])

            deployed_config.parameters[key] = value.rstrip("\n")

    def _remove_deployed_default_parameters_that_arent_passed(
        self,
        template_summary: dict,
        generated_config: StackConfiguration,
        deployed_config: StackConfiguration,
    ):
        deployed_config_default_map = self._get_parameter_default_map(template_summary)
        for parameter_key, default_value in deployed_config_default_map.items():
            # If the generated config defines that parameter, leave that in the deployed config
            # so we can see the diff.
            if parameter_key in generated_config.parameters:
                continue

            # But if the stack config is relying on the template's default value for this parameter,
            # remove that parameter from the deployed config... but only if its value is set to the
            # default value. If we don't do this, it will show up as a difference when it isn't.
            if deployed_config.parameters[parameter_key] == default_value:
                del deployed_config.parameters[parameter_key]

    def _get_parameter_default_map(self, template_summary: dict) -> Dict[str, str]:
        if template_summary is None:
            return {}

        parameters = template_summary["Parameters"]
        default_map = {}

        for parameter in parameters:
            key = parameter["ParameterKey"]
            value = self._handle_default_value(parameter)
            if value is not None:
                default_map[key] = value

        return default_map

    def _handle_default_value(self, parameter):
        default_value = parameter.get("DefaultValue")
        param_type = parameter["ParameterType"]

        if default_value is None:
            return None

        if parameter.get("NoEcho"):
            default_value = "****"
        elif "List" in param_type:
            # Eliminate whitespace around commas
            default_value = ",".join(
                value.strip() for value in default_value.split(",")
            )

        return default_value

    def _mask_no_echo_parameters(
        self, template_summary: dict, generated_config: StackConfiguration
    ):
        parameters = template_summary["Parameters"]

        for parameter in parameters:
            key = parameter["ParameterKey"]
            if parameter.get("NoEcho") and key in generated_config.parameters:
                generated_config.parameters[key] = self.NO_ECHO_REPLACEMENT

    def _generate_template(self, stack_actions: StackActions) -> str:
        return stack_actions.dump_template()

    def _get_deployed_template(
        self, stack_actions: StackActions, is_deployed: bool
    ) -> str:
        if is_deployed:
            return stack_actions.fetch_remote_template() or "{}"
        else:
            return "{}"

    @abstractmethod
    def compare_templates(self, deployed: str, generated: str) -> DiffType:
        """Implement this method to return the diff for the templates

        :param deployed: The stack template as it has been deployed
        :param generated: The stack template as it exists locally within Sceptre
        :return: The generated diff between the two
        """

    @abstractmethod
    def compare_stack_configurations(
        self, deployed: Optional[StackConfiguration], generated: StackConfiguration
    ) -> DiffType:
        """Implement this method to return the diff for the stack configurations.

        :param deployed: The StackConfiguration as it has been deployed. This MIGHT be None, if the
            stack has not been deployed.
        :param generated: The StackConfiguration as it exists locally within Sceptre
        :return: The generated diff between the two
        """


class DeepDiffStackDiffer(StackDiffer[deepdiff.DeepDiff]):
    """A StackDiffer that relies upon the DeepDiff library to produce the difference between the
    stack as it has been deployed onto CloudFormation and as it exists locally within Sceptre.

    This differ relies upon a recursive key/value comparison of Python data structures, indicating
    specific keys or values that have been added, removed, or altered between the two. Templates
    are read in as dictionaries and compared this way, so json or yaml formatting changes will not
    be reflected, only changes in value.
    """

    VERBOSITY_LEVEL_TO_INDICATE_CHANGED_VALUES = 2

    def __init__(
        self,
        show_no_echo=False,
        *,
        universal_template_loader: Callable[[str], Tuple[dict, str]] = cfn_flip.load,
    ):
        """Initializes a DeepDiffStackDiffer.

        :param show_no_echo: If True, local parameters passed that the template says are NoEcho
            parameters will be displayed; Otherwise, they will be masked in the diff.
        :param universal_template_loader: This should be a callable that can load either a json or
            yaml string and return a tuple where the first element is the loaded template and the
            second element is the template format (either "json" or "yaml")
        """
        super().__init__(show_no_echo)
        self.load_template = universal_template_loader

    def compare_stack_configurations(
        self,
        deployed: Optional[StackConfiguration],
        generated: StackConfiguration,
    ) -> deepdiff.DeepDiff:
        return deepdiff.DeepDiff(
            deployed,
            generated,
            verbose_level=self.VERBOSITY_LEVEL_TO_INDICATE_CHANGED_VALUES,
        )

    def compare_templates(self, deployed: str, generated: str) -> deepdiff.DeepDiff:
        # We don't actually care about the original formats here, since we only care about the
        # template VALUES.
        deployed_dict, _ = self.load_template(deployed)
        generated_dict, _ = self.load_template(generated)

        return deepdiff.DeepDiff(
            deployed_dict,
            generated_dict,
            verbose_level=self.VERBOSITY_LEVEL_TO_INDICATE_CHANGED_VALUES,
        )


class DifflibStackDiffer(StackDiffer[List[str]]):
    """A StackDiffer that uses difflib to produce a diff between the stack as it exists on AWS and
    the stack as it exists locally within Sceptre.

    Because difflib generates diffs off of lists of strings, both StackConfigurations and
    """

    def __init__(
        self,
        show_no_echo=False,
        *,
        universal_template_loader: Callable[[str], Tuple[dict, str]] = cfn_flip.load,
    ):
        """Initializes a DifflibStackDiffer.

        :param show_no_echo: If True, local parameters passed that the template says are NoEcho
            parameters will be displayed; Otherwise, they will be masked in the diff.
        :param universal_template_loader: This should be a callable that can load either a json or
            yaml string and return a tuple where the first element is the loaded template and the
            second element is the template format (either "json" or "yaml")
        """
        super().__init__(show_no_echo)
        self.load_template = universal_template_loader

    def compare_stack_configurations(
        self,
        deployed: Optional[StackConfiguration],
        generated: StackConfiguration,
    ) -> List[str]:
        if deployed is None:
            comparable_deployed = None
        else:
            comparable_deployed = self._make_stack_configuration_comparable(deployed)

        comparable_generated = self._make_stack_configuration_comparable(generated)
        deployed_string = cfn_flip.dump_yaml(comparable_deployed)
        generated_string = cfn_flip.dump_yaml(comparable_generated)
        return self._make_string_diff(deployed_string, generated_string)

    def _make_stack_configuration_comparable(
        self, config: Optional[StackConfiguration]
    ):
        as_dict = dict(config._asdict())
        return {
            key: value
            for key, value in as_dict.items()
            # stack_name isn't always going to be the same, otherwise we wouldn't be comparing them.
            # It's more confusing to have it in the diff output than to just remove it.
            if value not in (None, [], {}) and key != "stack_name"
        }

    def compare_templates(
        self,
        deployed: str,
        generated: str,
    ) -> List[str]:
        # Sometimes there might only be simple whitespace differences... which difflib will show but
        # are actually insignificant and "false positives". Also, it's POSSIBLE that the template
        # format might have changed, even if all the VALUES have stayed the same, so we'll read both
        # templates into dicts (regardless of their format) and we'll output both to the format of
        # the local template using identical serialization settings. This will truly enable comparison
        # of the actual values rather than other things that don't actually make a difference to
        # CloudFormation. If only the format/meaningless whitespace has changed, this will result in
        # there being no diff.
        deployed_dict, _ = self.load_template(deployed)
        generated_dict, generated_format = self.load_template(generated)
        dumpers = {"json": cfn_flip.dump_json, "yaml": cfn_flip.dump_yaml}
        deployed_reformatted = dumpers[generated_format](deployed_dict)
        generated_reformatted = dumpers[generated_format](generated_dict)

        return self._make_string_diff(deployed_reformatted, generated_reformatted)

    def _make_string_diff(self, deployed: str, generated: str) -> List[str]:
        diff_lines = difflib.unified_diff(
            deployed.splitlines(),
            generated.splitlines(),
            fromfile="deployed",
            tofile="generated",
            lineterm="",
        )
        return list(diff_lines)



# -*- coding: utf-8 -*-

import os

from sceptre.resolvers import Resolver


class EnvironmentVariable(Resolver):
    """
    Resolver for shell environment variables.

    :param argument: Name of the environment variable to return.
    :type argument: str
    """

    def __init__(self, *args, **kwargs):
        super(EnvironmentVariable, self).__init__(*args, **kwargs)

    def resolve(self):
        """
        Retrieves the value of a named environment variable.

        :returns: Value of the environment variable.
        :rtype: str
        """
        value = None
        if self.argument:
            value = os.environ.get(self.argument)
        return value



import click

from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions, simplify_change_set_description, write
from sceptre.plan.plan import SceptrePlan


@click.group(name="describe")
@click.pass_context
def describe_group(ctx):
    """
    Commands for describing attributes of stacks.
    """
    pass


@describe_group.command(name="change-set")
@click.argument("path")
@click.argument("change-set-name")
@click.option("-v", "--verbose", is_flag=True, help="Display verbose output.")
@click.pass_context
@catch_exceptions
def describe_change_set(ctx, path, change_set_name, verbose):
    """
    Describes the change set.
    \f

    :param path: Path to execute the command on.
    :type path: str
    :param change_set_name: Name of the Change Set to use.
    :type change_set_name: str
    :param verbose: A flag to display verbose output.
    :type verbose: bool
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        no_colour=ctx.obj.get("no_colour"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)

    responses = plan.describe_change_set(change_set_name)
    for response in responses.values():
        description = response
        if not verbose:
            description = simplify_change_set_description(description)
        write(description, context.output_format, context.no_colour)


@describe_group.command(name="policy")
@click.argument("path")
@click.pass_context
@catch_exceptions
def describe_policy(ctx, path):
    """
    Displays the stack policy used.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        no_colour=ctx.obj.get("no_colour"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)
    responses = plan.get_policy()
    for response in responses.values():
        write(response, context.output_format, context.no_colour)



# -*- coding: utf-8 -*-
import sceptre.template_handlers.helper as helper

from os import path
from pathlib import Path

from sceptre.exceptions import UnsupportedTemplateFileTypeError
from sceptre.template_handlers import TemplateHandler
from sceptre.helpers import normalise_path


class File(TemplateHandler):
    """
    Template handler that can load files from disk. Supports JSON, YAML, Jinja2 and Python.
    """

    def __init__(self, *args, **kwargs):
        super(File, self).__init__(*args, **kwargs)

    def schema(self):
        return {
            "type": "object",
            "properties": {
                "path": {"type": "string"},
            },
            "required": ["path"],
        }

    def handle(self):
        input_path = Path(self.arguments["path"])
        path = self._resolve_template_path(str(input_path))

        if input_path.suffix not in self.supported_template_extensions:
            raise UnsupportedTemplateFileTypeError(
                "Template has file extension %s. Only %s are supported.",
                input_path.suffix,
                ",".join(self.supported_template_extensions),
            )

        try:
            if input_path.suffix in self.standard_template_extensions:
                with open(path) as template_file:
                    return template_file.read()
            elif input_path.suffix in self.jinja_template_extensions:
                return helper.render_jinja_template(
                    path,
                    {"sceptre_user_data": self.sceptre_user_data},
                    self.stack_group_config.get("j2_environment", {}),
                )
            elif input_path.suffix in self.python_template_extensions:
                return helper.call_sceptre_handler(path, self.sceptre_user_data)
        except Exception as e:
            helper.print_template_traceback(path)
            raise e

    def _resolve_template_path(self, template_path):
        """
        Return the project_path joined to template_path as
        a string.

        Note that os.path.join defers to an absolute path
        if the input is absolute.
        """
        return path.join(
            self.stack_group_config["project_path"],
            "templates",
            normalise_path(template_path),
        )



import click

from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions
from sceptre.plan.plan import SceptrePlan


@click.command(name="set-policy", short_help="Sets Stack policy.")
@click.argument("path")
@click.argument("policy-file", required=False)
@click.option(
    "-b",
    "--built-in",
    type=click.Choice(["deny-all", "allow-all"]),
    help="Specify a built in stack policy.",
)
@click.pass_context
@catch_exceptions
def set_policy_command(ctx, path, policy_file, built_in):
    """
    Sets a specific Stack policy for either a file or using a built-in policy.
    \f

    :param path: Path to execute the command on.
    :type path: str
    :param policy_file: path to the AWS Policy file to use.
    :type policy_file: str
    :param built_in: the name of the built-in policy file to use.
    :type built_in: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )
    plan = SceptrePlan(context)

    if built_in == "deny-all":
        plan.lock()
    elif built_in == "allow-all":
        plan.unlock()
    else:
        plan.set_policy(policy_file)



import logging
import sys

from itertools import cycle
from functools import partial, wraps

from typing import Any, Optional
from pathlib import Path

import json
import click
import six
import yaml

from boto3.exceptions import Boto3Error
from botocore.exceptions import BotoCoreError, ClientError
from jinja2.exceptions import TemplateError

from sceptre.helpers import logging_level
from sceptre.exceptions import SceptreException
from sceptre.stack_status import StackStatus
from sceptre.stack_status_colourer import StackStatusColourer

logger = logging.getLogger(__name__)


def catch_exceptions(func):
    """
    Catches and simplifies expected errors thrown by sceptre.

    catch_exceptions should be used as a decorator.

    :param func: The function which may throw exceptions which should be
        simplified.
    :returns: The decorated function.
    """

    @wraps(func)
    def decorated(*args, **kwargs):
        """
        Invokes ``func``, catches expected errors, prints the error message and
        exits sceptre with a non-zero exit code. In debug mode, the original
        exception is re-raised to assist debugging.
        """
        try:
            return func(*args, **kwargs)
        except (
            SceptreException,
            BotoCoreError,
            ClientError,
            Boto3Error,
            TemplateError,
        ) as error:
            if logging_level() == logging.DEBUG:
                raise
            write(error)
            sys.exit(1)

    return decorated


def confirmation(command, ignore, command_path, change_set=None):
    if not ignore:
        msg = "Do you want to {} ".format(command)
        if change_set:
            msg = msg + "change set '{0}' for '{1}'".format(change_set, command_path)
        else:
            msg = msg + "'{0}'".format(command_path)
        click.confirm(msg, abort=True)


def write(
    var: Any,
    output_format: str = "json",
    no_colour: bool = True,
    file_path: Optional[Path] = None,
) -> None:
    """
    Writes ``var`` to stdout. If output_format is set to "json" or "yaml",
    write ``var`` as a JSON or YAML string.

    :param var: The object to print
    :param output_format: The format to print the output as. Allowed values: \
    "text", "json", "yaml"
    :param no_colour: Whether to colour stack statuses
    :param file_path: Optional path to a file to save the output
    """
    output = var

    if output_format == "json":
        output = _generate_json(var)
    if output_format == "yaml":
        output = _generate_yaml(var)
    if output_format == "text":
        output = _generate_text(var)

    if file_path:
        dir_path = file_path.parent
        dir_path.mkdir(parents=True, exist_ok=True)

        with open(file_path, "w") as f:
            f.write(output)

        return

    if not no_colour:
        stack_status_colourer = StackStatusColourer()
        output = stack_status_colourer.colour(str(output))

    click.echo(output)


def _generate_json(stream):
    encoder = CustomJsonEncoder(indent=4)
    if isinstance(stream, list):
        items = []
        for item in stream:
            try:
                if isinstance(item, dict):
                    items.append(item)
                else:
                    items.append(yaml.load(item, Loader=CfnYamlLoader))
            except Exception:
                print("An error occured writing the JSON object.")
        return encoder.encode(items)
    else:
        try:
            return encoder.encode(yaml.load(stream, Loader=CfnYamlLoader))
        except Exception:
            return encoder.encode(stream)


def _generate_yaml(stream):
    kwargs = {"default_flow_style": False, "explicit_start": True}

    if isinstance(stream, (list, set)):
        items = []
        for item in stream:
            try:
                if isinstance(item, dict):
                    items.append(yaml.safe_dump(item, **kwargs))
                else:
                    items.append(
                        yaml.safe_dump(yaml.load(item, Loader=CfnYamlLoader), **kwargs)
                    )
            except Exception:
                print("An error occured whilst writing the YAML object.")
        return yaml.safe_dump(
            [yaml.load(item, Loader=CfnYamlLoader) for item in items], **kwargs
        )

    elif isinstance(stream, dict):
        return yaml.dump(stream, **kwargs)

    else:
        try:
            return yaml.safe_loads(stream)
        except Exception:
            return stream


def _generate_text(stream):
    if isinstance(stream, list):
        items = []
        for item in stream:
            try:
                if isinstance(item, dict):
                    # use keys as headers, and add a blank row
                    if not items:
                        items = [["Stack"]]
                        items[0].extend(list(next(iter(*item.values()))))
                        items.append(["" for _ in range(len(items[0]))])
                    for k, v in item.items():
                        for r in item[k]:
                            row = [k]
                            row.extend(list(r.values()))
                            items.append(row)
                else:
                    items.append(item)
            except Exception:
                print("An error occured writing the text object.")
        col_widths = [max(len(c) for c in b) for b in zip(*items)]
        rows = []
        for row in items:
            rows.append(
                "".join([field for field, width in zip(row, cycle(col_widths))])
            )
        return "\n".join(rows)
    return stream


def setup_vars(var_file, var, merge_vars, debug, no_colour):
    """
    Handle --var-file and --var arguments before
    returning data for the user_variables as required
    by the ConfigReader and SceptreContext.

    :param var_file: the var_file list.
    :type var_file: List[Dict]
    :param var: the var list.
    :type var: List[str]
    :param merge_vars: Merge instead of
        overwrite duplicate keys.
    :type merge_vars: bool
    :param debug: debug mode.
    :type debug: bool
    :param no_colour: no_colour mode.
    :type no_colour: bool

    :returns: data for the user_variables.
    :rtype: Dict
    """
    logger = setup_logging(debug, no_colour)

    return_value = {}

    def _update_dict(variable):
        variable_key, variable_value = variable.split("=")
        keys = variable_key.split(".")

        def _nested_set(dic, keys, value):
            for key in keys[:-1]:
                dic = dic.setdefault(key, {})
            dic[keys[-1]] = value

        _nested_set(return_value, keys, variable_value)

    if var_file:
        for fh in var_file:
            parsed = yaml.safe_load(fh.read()) or {}

            if merge_vars:
                return_value = _deep_merge(parsed, return_value)
            else:
                return_value.update(parsed)

            # the rest of this block is for debug purposes only
            existing_keys = set(return_value.keys())
            new_keys = set(parsed.keys())
            overloaded_keys = existing_keys & new_keys  # intersection

            if overloaded_keys:
                message = "Duplicate variables encountered: "

                if merge_vars:
                    message += "{0}. Using values from: {1}.".format(
                        ", ".join(overloaded_keys), fh.name
                    )
                else:
                    message += "{0}. Performing deep merge, {1} wins.".format(
                        ", ".join(overloaded_keys), fh.name
                    )

                logger.debug(message)

    if var:
        # --var options overwrite --var-file options, unless a dict and --merge-vars.
        for variable in var:
            if isinstance(variable, dict) and merge_vars:
                return_value = _deep_merge(variable, return_value)
            else:
                _update_dict(variable)

    return return_value


def _deep_merge(source, destination):
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            _deep_merge(value, node)
        else:
            destination[key] = value

    return destination


def stack_status_exit_code(statuses):
    if not all(status == StackStatus.COMPLETE for status in statuses):
        return 1
    else:
        return 0


def setup_logging(debug, no_colour):
    """
    Sets up logging.

    By default, the python logging module is configured to push logs to stdout
    as long as their level is at least INFO. The log format is set to
    "[%(asctime)s] - %(name)s - %(message)s" and the date format is set to
    "%Y-%m-%d %H:%M:%S".

    After this function has run, modules should:

    .. code:: python

        import logging

        logging.getLogger(__name__).info("my log message")

    :param debug: A flag indication whether to turn on debug logging.
    :type debug: bool
    :no_colour: A flag to indicating whether to turn off coloured output.
    :type no_colour: bool
    :returns: A logger.
    :rtype: logging.Logger
    """
    if debug:
        sceptre_logging_level = logging.DEBUG
        logging.getLogger("botocore").setLevel(logging.INFO)
    else:
        sceptre_logging_level = logging.INFO
        # Silence botocore logs
        logging.getLogger("botocore").setLevel(logging.CRITICAL)

    formatter_class = logging.Formatter if no_colour else ColouredFormatter

    formatter = formatter_class(
        fmt="[%(asctime)s] - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )

    log_handler = logging.StreamHandler()
    log_handler.setFormatter(formatter)
    logger = logging.getLogger("sceptre")
    logger.addHandler(log_handler)
    logger.setLevel(sceptre_logging_level)
    return logger


def simplify_change_set_description(response):
    """
    Simplies the response from the AWS describe change set API.

    :param response: The original api response.
    :type response: dict
    :returns: A more concise description of the change set.
    :rtype: dict
    """
    if not response:
        return {"ChangeSetName": "ChangeSetNotFound"}

    desired_response_items = [
        "ChangeSetName",
        "CreationTime",
        "ExecutionStatus",
        "StackName",
        "Status",
        "StatusReason",
    ]
    desired_resource_changes = [
        "Action",
        "LogicalResourceId",
        "PhysicalResourceId",
        "Replacement",
        "ResourceType",
        "Scope",
    ]
    formatted_response = {
        k: v for k, v in response.items() if k in desired_response_items
    }
    formatted_response["Changes"] = [
        {
            "ResourceChange": {
                k: v
                for k, v in change["ResourceChange"].items()
                if k in desired_resource_changes
            }
        }
        for change in response["Changes"]
    ]
    return formatted_response


def deserialize_json_properties(value):
    if isinstance(value, str):
        is_json = (value.startswith("{") and value.endswith("}")) or (
            value.startswith("[") and value.endswith("]")
        )
        if is_json:
            return json.loads(value)
        return value
    if isinstance(value, dict):
        return {key: deserialize_json_properties(val) for key, val in value.items()}
    if isinstance(value, list):
        return [deserialize_json_properties(item) for item in value]
    return value


class ColouredFormatter(logging.Formatter):
    """
    ColouredFormatter add colours to all stack statuses that appear in log
    messages.
    """

    stack_status_colourer = StackStatusColourer()

    def format(self, record):
        """
        Colours and returns all stack statuses in ``record``.

        :param record: The log item to format.
        :type record: str
        :returns: str
        """
        response = super(ColouredFormatter, self).format(record)
        coloured_response = self.stack_status_colourer.colour(response)
        return coloured_response


class CustomJsonEncoder(json.JSONEncoder):
    """
    CustomJsonEncoder is a JSONEncoder which encodes all items as JSON by
    calling their __str__() method.
    """

    def default(self, item):
        """
        Returns stringified version of item.

        :param item: An arbitrary object to stringify.
        :type item: object
        :returns: The stringified object.
        :rtype: str
        """
        return str(item)


CFN_FNS = [
    "And",
    "Base64",
    "Cidr",
    "Equals",
    "FindInMap",
    "GetAtt",
    "GetAZs",
    "If",
    "ImportValue",
    "Join",
    "Not",
    "Or",
    "Select",
    "Split",
    "Sub",
    "Transform",
]

CFN_TAGS = [
    "Condition",
    "Ref",
]


def _getatt_constructor(loader, node):
    if isinstance(node.value, six.text_type):
        return node.value.split(".", 1)
    elif isinstance(node.value, list):
        seq = loader.construct_sequence(node)
        for item in seq:
            if not isinstance(item, six.text_type):
                raise ValueError("Fn::GetAtt does not support complex datastructures")
        return seq
    else:
        raise ValueError("Fn::GetAtt only supports string or list values")


def _tag_constructor(loader, tag_suffix, node):
    if tag_suffix not in CFN_FNS and tag_suffix not in CFN_TAGS:
        raise ValueError(
            "Bad tag: !{tag_suffix}. Supported tags are: "
            "{supported_tags}".format(
                tag_suffix=tag_suffix,
                supported_tags=", ".join(sorted(CFN_TAGS + CFN_FNS)),
            )
        )

    if tag_suffix in CFN_FNS:
        tag_suffix = "Fn::{tag_suffix}".format(tag_suffix=tag_suffix)

    data = {}
    yield data

    if tag_suffix == "Fn::GetAtt":
        constructor = partial(_getatt_constructor, (loader,))
    elif isinstance(node, yaml.ScalarNode):
        constructor = loader.construct_scalar
    elif isinstance(node, yaml.SequenceNode):
        constructor = loader.construct_sequence
    elif isinstance(node, yaml.MappingNode):
        constructor = loader.construct_mapping

    data[tag_suffix] = constructor(node)


class CfnYamlLoader(yaml.SafeLoader):
    pass


CfnYamlLoader.add_multi_constructor("!", _tag_constructor)



# -*- coding: utf-8 -*-

"""
sceptre.plan.plan

This module implements a SceptrePlan, which is responsible for holding all
nessessary information for a command to execute.
"""
import functools
import itertools
import pathlib

from os import path, walk
from typing import Dict, List, Set, Callable, Iterable, Optional

from sceptre.config.graph import StackGraph
from sceptre.config.reader import ConfigReader
from sceptre.context import SceptreContext
from sceptre.diffing.stack_differ import StackDiff
from sceptre.exceptions import ConfigFileNotFoundError
from sceptre.helpers import sceptreise_path
from sceptre.plan.executor import SceptrePlanExecutor
from sceptre.stack import Stack


def require_resolved(func) -> Callable:
    @functools.wraps(func)
    def wrapped(self: "SceptrePlan", *args, **kwargs):
        if self.launch_order is None:
            raise RuntimeError(f"You cannot call {func.__name__}() before resolve().")
        return func(self, *args, **kwargs)

    return wrapped


class SceptrePlan(object):
    def __init__(self, context: SceptreContext):
        """
        Intialises a SceptrePlan and generates the Stacks, StackGraph and
        launch order of required.

        :param context: A SceptreContext
        """
        self.context = context
        self.command = None
        self.reverse = None
        self.launch_order: Optional[List[Set[Stack]]] = None

        self.config_reader = ConfigReader(context)
        all_stacks, command_stacks = self.config_reader.construct_stacks()
        self.graph = StackGraph(all_stacks)
        self.command_stacks = command_stacks

    @require_resolved
    def _execute(self, *args):
        executor = SceptrePlanExecutor(self.command, self.launch_order)
        return executor.execute(*args)

    def _raise_no_launch_order_error(self):
        MAX_VALID_STACK_PATH_COUNT = 10

        error_text = f"No stacks detected from the given path '{sceptreise_path(self.context.command_path)}'."
        command_path_parent = pathlib.PurePath(self.context.command_path).parent

        all_paths = self._valid_stack_paths()
        filtered_valid_stack_paths = [
            p for p in all_paths if pathlib.PurePath(p).parent == command_path_parent
        ] or all_paths
        if (
            filtered_valid_stack_paths
            and len(filtered_valid_stack_paths) < MAX_VALID_STACK_PATH_COUNT
        ):
            error_text += (
                f" Valid stack paths under '{sceptreise_path(str(command_path_parent))}' "
                + f"are: {filtered_valid_stack_paths}"
            )
        raise ConfigFileNotFoundError(error_text)

    def _generate_launch_order(self, reverse=False) -> List[Set[Stack]]:
        if self.context.ignore_dependencies:
            return [self.command_stacks]

        graph = self.graph.filtered(self.command_stacks, reverse)

        launch_order = []
        while graph.graph:
            batch = set()
            for stack in graph:
                if graph.count_dependencies(stack) == 0:
                    batch.add(stack)
            launch_order.append(batch)

            for stack in batch:
                graph.remove_stack(stack)

        if not launch_order:
            self._raise_no_launch_order_error()

        return launch_order

    @require_resolved
    def __iter__(self) -> Iterable[Stack]:
        """Iterates the stacks in the launch_order"""
        # We cast it to list so it's "frozen" in time, in case the launch order is modified
        # while iterating.
        yield from list(itertools.chain.from_iterable(self.launch_order))

    @require_resolved
    def remove_stack_from_plan(self, stack: Stack):
        for batch in self.launch_order:
            if stack in batch:
                batch.remove(stack)
                return

    @require_resolved
    def filter(self, predicate: Callable[[Stack], bool]):
        """Filters the plan's resolved launch_order to remove specific stacks.

        :param predicate: This callable should take a single Stack and return True if it should stay
            in the launch_order or False if it should be filtered out.
        """
        for stack in self:
            if not predicate(stack):
                self.remove_stack_from_plan(stack)

    def resolve(self, command, reverse=False):
        if command == self.command and reverse == self.reverse:
            return

        self.command = command
        self.reverse = reverse
        self.launch_order = self._generate_launch_order(reverse)

    def template(self, *args):
        """
        Returns the CloudFormation Template used to create the Stack.

        :returns: A dictionary of Stacks and their templates.
        :rtype: dict
        """
        self.resolve(command=self.template.__name__)
        return self._execute(*args)

    def create(self, *args):
        """
        Creates the Stack.

        :returns: A dictionary of Stacks and their status.
        :rtype: dict
        """
        self.resolve(command=self.create.__name__)
        return self._execute(*args)

    def update(self, *args):
        """
        Updates the Stack.

        :returns: A dictionary of Stacks and their status.
        :rtype: dict
        """
        self.resolve(command=self.update.__name__)
        return self._execute(*args)

    def cancel_stack_update(self, *args):
        """
        Cancels a Stack update.

        :returns: A dictionary of Stacks and their cancelled statuses.
        :rtype: dict
        """
        self.resolve(command=self.cancel_stack_update.__name__)
        return self._execute(*args)

    def launch(self, *args):
        """
        Launches the Stack.

        If the Stack status is create_failed or rollback_complete, the
        Stack is deleted. Launch then tries to create or update the Stack,
        depending if it already exists. If there are no updates to be
        performed, launch exits gracefully.

        :returns: A dictionary of Stacks and their status.
        :rtype: dict
        """
        self.resolve(command=self.launch.__name__)
        return self._execute(*args)

    def delete(self, *args):
        """
        Deletes the Stack.

        :returns: A dictionary of Stacks and their status.
        :rtype: dict
        """
        self.resolve(command=self.delete.__name__, reverse=True)
        return self._execute(*args)

    def lock(self, *args):
        """
        Locks the Stack by applying a deny all updates Stack policy.

        :returns: A dictionary of Stacks
        :rtype: dict
        """
        self.resolve(command=self.lock.__name__)
        return self._execute(*args)

    def unlock(self, *args):
        """
        Unlocks the Stack by applying an allow all updates Stack policy.

        :returns: A dictionary of Stacks
        :rtype: dict
        """
        self.resolve(command=self.unlock.__name__)
        return self._execute(*args)

    def describe(self, *args):
        """
        Returns the a description of the Stack.

        :returns: A dictionary of Stacks and their description.
        :rtype: dict
        """
        self.resolve(command=self.describe.__name__)
        return self._execute(*args)

    def describe_events(self, *args):
        """
        Returns a dictionary contianing the Stack events.

        :returns: A dictionary of Stacks and their CloudFormation events.
        :rtype: dict
        """
        self.resolve(command=self.describe_events.__name__)
        return self._execute(*args)

    def describe_resources(self, *args):
        """
        Returns the logical and physical resource IDs of the Stack's resources.

        :returns: A dictionary of Stacks and their resources.
        :rtype: dict
        """
        self.resolve(command=self.describe_resources.__name__)
        return self._execute(*args)

    def describe_outputs(self, *args):
        """
        Returns a list of Stack outputs.

        :returns: A dictionary of Stacks and their outputs.
        :rtype: dict
        """
        self.resolve(command=self.describe_outputs.__name__)
        return self._execute(*args)

    def continue_update_rollback(self, *args):
        """
        Rolls back a Stack in the UPDATE_ROLLBACK_FAILED state to
        UPDATE_ROLLBACK_COMPLETE.

        :returns: A dictionary of Stacks
        :rtype: dict
        """
        self.resolve(command=self.continue_update_rollback.__name__)
        return self._execute(*args)

    def set_policy(self, *args):
        """
        Applies a Stack policy.

        :param policy_path: the path of json file containing a aws policy
        :type policy_path: str
        :returns: A dictionary of Stacks
        :rtype: dict
        """
        self.resolve(command=self.set_policy.__name__)
        return self._execute(*args)

    def get_policy(self, *args):
        """
        Returns a Stack's policy.

        :returns: A dictionary of Stacks and their Stack policy.
        :rtype: dict
        """
        self.resolve(command=self.get_policy.__name__)
        return self._execute(*args)

    def create_change_set(self, *args):
        """
        Creates a Change Set with the name ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: A dictionary of Stacks
        :rtype: dict
        """
        self.resolve(command=self.create_change_set.__name__)
        return self._execute(*args)

    def delete_change_set(self, *args):
        """
        Deletes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: A dictionary of Stacks
        :rtype: dict
        """
        self.resolve(command=self.delete_change_set.__name__)
        return self._execute(*args)

    def describe_change_set(self, *args):
        """
        Describes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: A dictionary of Stacks and their Change Set description.
        :rtype: dict
        """
        self.resolve(command=self.describe_change_set.__name__)
        return self._execute(*args)

    def execute_change_set(self, *args):
        """
        Executes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: A dictionary of Stacks and their status.
        :rtype: dict
        """
        self.resolve(command=self.execute_change_set.__name__)
        return self._execute(*args)

    def list_change_sets(self, *args):
        """
        Lists the Stack's Change Sets.

        :returns: TA dictionary of Stacks and their Change Sets.
        :rtype: dict
        """
        self.resolve(command=self.list_change_sets.__name__)
        return self._execute(*args)

    def get_status(self, *args):
        """
        Returns the Stack's status.

        :returns: A dictionary of Stacks and their status.
        :rtype: dict
        :raises: sceptre.exceptions.StackDoesNotExistError
        """
        self.resolve(command=self.get_status.__name__)
        return self._execute(*args)

    def wait_for_cs_completion(self, *args):
        """
        Waits while the Stack Change Set status is "pending".

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :rtype: dict
        :rtype: sceptre.stack_status.StackChangeSetStatus
        """
        self.resolve(command=self.wait_for_cs_completion.__name__)
        return self._execute(*args)

    def validate(self, *args):
        """
        Validates the Stack's CloudFormation template.

        Raises an error if the Template is invalid.

        :returns: A dictionary of Stacks and their template validation information.
        :rtype: dict
        :raises: botocore.exceptions.ClientError
        """
        self.resolve(command=self.validate.__name__)
        return self._execute(*args)

    def estimate_cost(self, *args):
        """
        Estimates a Stack's cost.

        :returns: A dictionary of Stacks and their estimated costs.
        :rtype: dict
        :raises: botocore.exceptions.ClientError
        """
        self.resolve(command=self.estimate_cost.__name__)
        return self._execute(*args)

    def generate(self, *args):
        """
        Returns a generated Template for a given Stack. An alias for
        dump_template for historical reasons.

        :returns: A dictionary of Stacks and their template body.
        :rtype: dict
        """
        self.resolve(command=self.generate.__name__)
        return self._execute(*args)

    def _valid_stack_paths(self):
        return [
            sceptreise_path(
                path.relpath(path.join(dirpath, f), self.context.config_path)
            )
            for dirpath, dirnames, files in walk(self.context.config_path)
            for f in files
            if not f.endswith(self.context.config_file)
        ]

    def fetch_remote_template(self, *args):
        """
        Returns a generated Template for a given Stack

        :returns: A list of Stacks and their template body.
        :rtype: List[str]
        """
        self.resolve(command=self.fetch_remote_template.__name__)
        return self._execute(*args)

    def diff(self, *args) -> Dict[Stack, StackDiff]:
        """
        Show diffs between the running and generated stack.

        :returns: A dict where the keys are Stack objects and the values are StackDiffs.
        """
        self.resolve(command=self.diff.__name__)
        return self._execute(*args)

    def drift_detect(self, *args) -> Dict[Stack, str]:
        """
        Show drift detection status of a stack.

        :returns: A list of detected drift against running stacks.
        """
        self.resolve(command=self.drift_detect.__name__)
        return self._execute(*args)

    def drift_show(self, *args) -> Dict[Stack, str]:
        """
        Show stack drift for a running stack.

        :returns: A list of detected drift against running stacks.
        """
        self.resolve(command=self.drift_show.__name__)
        return self._execute(*args)

    def dump_config(self, *args):
        """
        Dump the config for a stack.
        """
        self.resolve(command=self.dump_config.__name__)
        return self._execute(*args)

    def dump_template(self, *args):
        """
        Dump the template for a stack. An alias
        for generate for historical reasons.
        """
        self.resolve(command=self.dump_template.__name__)
        return self._execute(*args)



# -*- coding: utf-8 -*-

"""
sceptre.template

This module implements a Template class, which stores a CloudFormation template
and implements methods for uploading it to S3.
"""

import logging
import threading
import botocore
import sys

import sceptre.helpers

from sceptre.exceptions import TemplateHandlerNotFoundError
from sceptre.logging import StackLoggerAdapter


class Template(object):
    """
    Template represents an AWS CloudFormation template. It is responsible for
    loading, storing and optionally uploading local templates for use by
    CloudFormation.

    :param name: The name of the template. Should be safe to use in filenames and not contain path segments.
    :type name: str

    :param handler_config: The configuration for a Template handler. Must contain a `type`.
    :type handler_config: dict

    :param sceptre_user_data: A dictionary of arbitrary data to be passed to\
            a handler function in an external Python script.
    :type sceptre_user_data: dict

    :param stack_group_config: The StackGroup config for the Stack.
    :type stack_group_config: dict

    :param connection_manager:
    :type connection_manager: sceptre.connection_manager.ConnectionManager

    :param s3_details:
    :type s3_details: dict
    """

    _boto_s3_lock = threading.Lock()

    def __init__(
        self,
        name,
        handler_config,
        sceptre_user_data,
        stack_group_config,
        connection_manager=None,
        s3_details=None,
    ):
        self.logger = StackLoggerAdapter(logging.getLogger(__name__), name)
        self.name = name
        self.handler_config = handler_config
        if self.handler_config is not None and self.handler_config.get("type") is None:
            self.handler_config["type"] = "file"
        self.sceptre_user_data = sceptre_user_data
        self.stack_group_config = stack_group_config
        self.connection_manager = connection_manager
        self.s3_details = s3_details

        self._registry = None
        self._body = None

    def __repr__(self):
        return sceptre.helpers.gen_repr(
            self,
            class_label="sceptre.template.Template",
            attributes=["name", "handler_config", "sceptre_user_data", "s3_details"],
        )

    @property
    def body(self):
        """
        Represents body of the CloudFormation template.

        :returns: The body of the CloudFormation template.
        :rtype: str
        """
        if self._body is None:
            type = self.handler_config.get("type")
            handler_class = self._get_handler_of_type(type)
            handler = handler_class(
                name=self.name,
                arguments={k: v for k, v in self.handler_config.items() if k != "type"},
                sceptre_user_data=self.sceptre_user_data,
                connection_manager=self.connection_manager,
                stack_group_config=self.stack_group_config,
            )
            handler.validate()
            body = handler.handle()
            if isinstance(body, bytes):
                body = body.decode("utf-8")
            if not str(body).startswith("---"):
                body = "---\n{}".format(body)
            self._body = body

        return self._body

    def upload_to_s3(self):
        """
        Uploads the template to ``bucket_name`` and returns its URL.

        The Template is uploaded with the ``bucket_key``.

        :returns: The URL of the Template object in S3.
        :rtype: str
        :raises: botocore.exceptions.ClientError

        """
        self.logger.debug("%s - Uploading template to S3...", self.name)

        with self._boto_s3_lock:
            if not self._bucket_exists():
                self._create_bucket()

        # Remove any leading or trailing slashes the user may have added.
        bucket_name = self.s3_details["bucket_name"]
        bucket_key = self.s3_details["bucket_key"]
        bucket_region = self._bucket_region(bucket_name)

        self.logger.debug(
            "%s - Uploading template to: 's3://%s/%s'",
            self.name,
            bucket_name,
            bucket_key,
        )
        self.connection_manager.call(
            service="s3",
            command="put_object",
            kwargs={
                "Bucket": bucket_name,
                "Key": bucket_key,
                "Body": self.body,
                "ServerSideEncryption": "AES256",
            },
        )

        url = "https://{}.s3.{}.amazonaws.{}/{}".format(
            bucket_name,
            bucket_region,
            self._domain_from_region(bucket_region),
            bucket_key,
        )

        self.logger.debug("%s - Template URL: '%s'", self.name, url)

        return url

    def _bucket_exists(self):
        """
        Checks if the bucket ``bucket_name`` exists.

        :returns: Boolean whether the bucket exists
        :rtype: bool
        :raises: botocore.exception.ClientError

        """
        bucket_name = self.s3_details["bucket_name"]
        self.logger.debug(
            "%s - Attempting to find template bucket '%s'", self.name, bucket_name
        )
        try:
            self.connection_manager.call(
                service="s3", command="head_bucket", kwargs={"Bucket": bucket_name}
            )
        except botocore.exceptions.ClientError as exp:
            if exp.response["Error"]["Message"] == "Not Found":
                self.logger.debug("%s - %s bucket not found.", self.name, bucket_name)
                return False
            else:
                raise
        self.logger.debug("%s - Found template bucket '%s'", self.name, bucket_name)
        return True

    def _create_bucket(self):
        """
        Create the s3 bucket ``bucket_name``.

        :raises: botocore.exception.ClientError

        """
        bucket_name = self.s3_details["bucket_name"]

        self.logger.debug("%s - Creating new bucket '%s'", self.name, bucket_name)

        if self.connection_manager.region == "us-east-1":
            self.connection_manager.call(
                service="s3", command="create_bucket", kwargs={"Bucket": bucket_name}
            )
        else:
            self.connection_manager.call(
                service="s3",
                command="create_bucket",
                kwargs={
                    "Bucket": bucket_name,
                    "CreateBucketConfiguration": {
                        "LocationConstraint": self.connection_manager.region
                    },
                },
            )

    def get_boto_call_parameter(self):
        """
        Returns the CloudFormation template location.

        Uploads the template to S3 and returns the object's URL, or returns
        the template itself.

        :returns: The boto call parameter for the template.
        :rtype: dict
        """
        # If bucket_name is set to None, it should be ignored and not uploaded.
        if self.s3_details and self.s3_details.get("bucket_name"):
            url = self.upload_to_s3()
            return {"TemplateURL": url}
        else:
            return {"TemplateBody": self.body}

    def _bucket_region(self, bucket_name):
        region = self.connection_manager.call(
            service="s3", command="get_bucket_location", kwargs={"Bucket": bucket_name}
        ).get("LocationConstraint")
        return region if region else "us-east-1"

    @staticmethod
    def _domain_from_region(region):
        return "com.cn" if region.startswith("cn-") else "com"

    @staticmethod
    def _iterate_entry_points(group, name):
        """
        Helper to determine whether to use pkg_resources or importlib.metadata.
        https://docs.python.org/3/library/importlib.metadata.html
        """
        if sys.version_info < (3, 10):
            from pkg_resources import iter_entry_points

            return iter_entry_points(group, name)
        else:
            from importlib.metadata import entry_points

            return entry_points(group=group, name=name)

    def _get_handler_of_type(self, type):
        """
        Gets a TemplateHandler type from the registry that can be used to get a string
        representation of a CloudFormation template.
        :param type: The type of Template Handler to load
        :type type: str
        :return: Instantiated TemplateHandler
        :rtype: class
        """
        if not self._registry:
            self._registry = {}

            for entry_point in self._iterate_entry_points(
                "sceptre.template_handlers", type
            ):
                self._registry[entry_point.name] = entry_point.load()

        if type not in self._registry:
            raise TemplateHandlerNotFoundError(
                'Handler of type "{0}" not found'.format(type)
            )

        return self._registry[type]



from sceptre.resolvers import Resolver


class Split(Resolver):
    """This resolver will split a value on a given delimiter string. This is great when combining with the
    ``!select`` resolver. This function works the same as CloudFormation's ``!Split`` intrinsic function.

    Note: The return value of this resolver is a *list*, not a string. This will not work to set Stack
    configurations that expect strings, but it WILL work to set Stack configurations that expect lists.

    The argument for this resolver should be a list with two elements: (1) The delimiter to split on and
    (2) a string to split.

    Example:
       notifications: !split
         - ";"
         - !stack_output my/sns/topics.yaml::SemicolonDelimitedArns
    """

    def resolve(self):
        error_message = (
            "The argument to !split must be a two-element list, where the first element is the "
            "string to split on and the second element string to split."
        )
        if (
            not isinstance(self.argument, list)
            or len(self.argument) != 2
            or not all(isinstance(a, str) for a in self.argument)
        ):
            self.raise_invalid_argument_error(error_message)

        split_on, split_string = self.argument
        return split_string.split(split_on)



import click

from typing import Optional
from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions, confirmation
from sceptre.plan.plan import SceptrePlan
from sceptre.cli.helpers import stack_status_exit_code


@click.command(name="create", short_help="Creates a stack or a change set.")
@click.argument("path")
@click.argument("change-set-name", required=False)
@click.option("-y", "--yes", is_flag=True, help="Assume yes to all questions.")
@click.option(
    "--disable-rollback/--enable-rollback",
    default=None,
    help="Disable or enable the cloudformation automatic rollback",
)
@click.pass_context
@catch_exceptions
def create_command(ctx, path, change_set_name, yes, disable_rollback: Optional[bool]):
    """
    Creates a stack for a given config PATH. Or if CHANGE_SET_NAME is specified
    creates a change set for stack in PATH.
    \f

    :param path: Path to a Stack or StackGroup
    :type path: str
    :param change_set_name: A name of the Change Set - optional
    :type change_set_name: str
    :param yes: A flag to assume yes to all questions.
    :type yes: bool
    :param disable_rollback: A flag to disable cloudformation rollback.
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    action = "create"
    plan = SceptrePlan(context)

    if change_set_name:
        confirmation(action, yes, change_set=change_set_name, command_path=path)
        plan.create_change_set(change_set_name)
    else:
        confirmation(action, yes, command_path=path)
        responses = plan.create()
        exit(stack_status_exit_code(responses.values()))



# -*- coding: utf-8 -*-
import abc
import logging

import six
from jsonschema import validate, ValidationError

from sceptre.exceptions import TemplateHandlerArgumentsInvalidError
from sceptre.logging import StackLoggerAdapter


@six.add_metaclass(abc.ABCMeta)
class TemplateHandler:
    """
    TemplateHandler is an abstract base class that should be inherited
    by all Template Handlers.

    :param name: Name of the template
    :type name: str

    :param arguments: The arguments of the template handler
    :type arguments: dict

    :param sceptre_user_data: Sceptre user data in stack config
    :type sceptre_user_data: dict

    :param connection_manager: Connection manager used to call AWS
    :type connection_manager: sceptre.connection_manager.ConnectionManager

    :param stack_group_config: The Stack group config to use as defaults.
    :type stack_group_config: dict
    """

    __metaclass__ = abc.ABCMeta

    standard_template_extensions = [".json", ".yaml", ".template"]
    jinja_template_extensions = [".j2"]
    python_template_extensions = [".py"]
    supported_template_extensions = (
        standard_template_extensions
        + jinja_template_extensions
        + python_template_extensions
    )

    def __init__(
        self,
        name,
        arguments=None,
        sceptre_user_data=None,
        connection_manager=None,
        stack_group_config=None,
    ):
        self.logger = StackLoggerAdapter(logging.getLogger(__name__), name)
        self.name = name
        self.arguments = arguments
        self.sceptre_user_data = sceptre_user_data
        self.connection_manager = connection_manager

        if stack_group_config is None:
            stack_group_config = {}
        self.stack_group_config = stack_group_config

    @abc.abstractmethod
    def schema(self):
        """
        Returns the schema for the arguments of this Template Resolver. This will
        be used to validate that the arguments passed in the stack config are what
        the Template Handler expects.
        :return: JSON schema that can be validated
        :rtype: object
        """
        pass

    @abc.abstractmethod
    def handle(self):
        """
        An abstract method which must be overwritten by all inheriting classes.
        This method is called to retrieve the template.
        Implementation of this method in subclasses must return a string that
        can be interpreted by Sceptre (CloudFormation YAML / JSON, Jinja or Python)
        """
        pass  # pragma: no cover

    def validate(self):
        """
        Validates if the current arguments are correct according to the schema. If this
        does not raise an exception, the template handler's arguments are valid.
        """
        try:
            validate(instance=self.arguments, schema=self.schema())
        except ValidationError as e:
            raise TemplateHandlerArgumentsInvalidError(e)



from sceptre.resolvers import Resolver


class Select(Resolver):
    """This resolver allows you to select a specific index from a list of items or a specific key
    from a dict.. This is great for combining with the ``!split`` resolver to obtain part of a
    string. This function works almost the same as CloudFormation's ``!Select`` intrinsic function,
    **except (1) you can use this with negative indexes to select with a reverse index** and (2)
    you can select keys from a dict.

    The argument for this resolver should be a list with two elements: (1) A numerical index or
    string key and (2) a list or dict of items to select out of. If the index is negative,
    it will select from the end of the list. For example, "-1" would select the last element and
    "-2" would select the second-to-last element.

    Example:

       parameters:
         # This selects the last element after you split the connection string on "/"
         DatabaseName: !select
           - -1
           - !split ["/", !stack_output my/database/stack.yaml::ConnectionString]
    """

    def resolve(self):
        error_message = (
            "The argument to !select must be a two-element list, where the first element is the "
            "index or key to select with and the second element is the list or dict to select from."
        )
        if not isinstance(self.argument, list) or len(self.argument) != 2:
            self.raise_invalid_argument_error(error_message)

        index, items = self.argument
        if not isinstance(items, (dict, list)):
            self.raise_invalid_argument_error(error_message)

        try:
            return items[index]
        except (TypeError, KeyError, IndexError) as e:
            self.raise_invalid_argument_error(
                f"Could not select with index/key {index}: {e}", e
            )



# -*- coding: utf-8 -*-

"""
sceptre.config.graph

This module implements a StackGraph, which is represented as a directed
acyclic graph of a Stack's dependencies.
"""

import logging
from typing import List

import networkx as nx
from sceptre.exceptions import CircularDependenciesError
from sceptre.stack import Stack


class StackGraph(object):
    """
    A Directed Acyclic Graph representing the relationship between a Stack
    and its dependencies. Responsible for initalising the graph based on a set
    of Stacks.
    """

    def __init__(self, stacks):
        """
        Initialises a StackGraph based on a `set` of Stacks.

        :param stacks: A set of Stacks.
        :type stacks: set
        """
        self.logger = logging.getLogger(__name__)
        self.graph = nx.DiGraph()
        self._generate_graph(stacks)

    def __repr__(self):
        return str(nx.convert.to_dict_of_lists(self.graph))

    def __iter__(self):
        return self.graph.__iter__()

    def filtered(self, source_stacks, reverse=False):
        graph = (nx.reverse if reverse else nx.DiGraph)(self.graph)

        relevant = set(source_stacks)
        for stack in source_stacks:
            relevant |= nx.algorithms.dag.ancestors(graph, stack)
        graph.remove_nodes_from({stack for stack in graph if stack not in relevant})

        filtered = StackGraph(set())
        filtered.graph = graph

        return filtered

    def count_dependencies(self, stack):
        """
        Returns the number of incoming edges a given Stack has in the
        StackGraph. The number of incoming edge also represents the number
        of Stacks that depend on the given Stack.
        """
        return self.graph.in_degree(stack)

    def remove_stack(self, stack):
        """
        Removes a Stack from the StackGraph. This operation will also remove
        all adjecent edges that represent a 'depends on' relationship with
        other Stacks.
        """
        return self.graph.remove_node(stack)

    def _generate_graph(self, stacks):
        """
        Generates the graph for the StackGraph object.

        :param stacks: A set of Stacks
        :type stacks: set
        """

        for stack in stacks:
            self._generate_edges(stack, stack.dependencies)
        self.graph.remove_edges_from(nx.selfloop_edges(self.graph))

    def _generate_edges(self, stack: Stack, dependencies: List[Stack]):
        """
        Adds edges to the graph based on a list of dependencies that are
        generated from the initial stack config. Each of the paths
        in the inital_dependency_paths list are a depency that the inital
        Stack config depends on.

        :param stack: A Sceptre Stack
        :param dependencies: a collection of dependency paths
        """
        self.logger.debug(f"Generate dependencies for stack {stack}")
        for dependency in set(dependencies):
            self.graph.add_edge(dependency, stack)
            try:
                cycle = nx.find_cycle(self.graph, orientation="original")
                cycle_str = ", ".join([f"{edge[0]} -> {edge[1]}" for edge in cycle])
                raise CircularDependenciesError(
                    f"Dependency cycle detected: {cycle_str}"
                )
            except nx.NetworkXNoCycle:
                pass  # No cycle, continue
            self.logger.debug(f"  Added dependency: {dependency}")

        if not dependencies:
            self.graph.add_node(stack)



# -*- coding: utf-8 -*-

from contextlib import contextmanager
from datetime import datetime
from os import sep
from typing import Optional, Any, List, Tuple, Union

import dateutil.parser
import deprecation
import logging
import tempfile

from sceptre.exceptions import PathConversionError
from sceptre import __version__


def logging_level():
    """
    Return the logging level.
    """
    logger = logging.getLogger(__name__)
    return logger.getEffectiveLevel()


def write_debug_file(content: str, prefix: str) -> str:
    """
    Write some content to a temp file for debug purposes.

    :param content: the file content to write.
    :returns: the full path to the temp file.
    """
    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, prefix=prefix
    ) as temp_file:
        temp_file.write(content)
        temp_file.flush()

    return temp_file.name


def get_external_stack_name(project_code, stack_name):
    """
    Returns the name given to a stack in CloudFormation.
    :param project_code: The project code, as defined in config.yaml.
    :type project_code: str
    :param stack_name: The name of the stack.
    :type stack_name: str
    :returns: The name given to the stack in CloudFormation.
    :rtype: str
    """
    return "-".join([project_code, stack_name.replace("/", "-")])


def mask_key(key):
    """
    Returns an masked version of ``key``.

    Returned version has all but the last four characters are replaced with the
    character "*".

    :param key: The string to mask.
    :type key: str
    :returns: An masked version of the key
    :rtype: str
    """
    num_mask_chars = len(key) - 4

    return "".join(["*" if i < num_mask_chars else c for i, c in enumerate(key)])


def _call_func_on_values(func, attr, cls):
    """
    Searches through dictionary or list for objects of type `cls` and calls the
    supplied function `func`. Supports nested dictionaries and lists.
    Does not detect objects used as keys in dictionaries.

    :param attr: A dictionary or list to search through.
    :type attr: dict or list
    :return: The dictionary or list structure.
    :rtype: dict or list
    """

    def func_on_instance(key):
        if isinstance(value, cls):
            func(attr, key, value)
        elif isinstance(value, list) or isinstance(value, dict):
            _call_func_on_values(func, value, cls)

    if isinstance(attr, dict):
        for key, value in attr.items():
            func_on_instance(key)
    elif isinstance(attr, list):
        for index, value in enumerate(attr):
            func_on_instance(index)
    return attr


Container = Union[list, dict]
Key = Union[str, int]


def delete_keys_from_containers(keys_to_delete: List[Tuple[Container, Key]]):
    """Removes the indicated keys/indexes from their paired containers."""
    list_items_to_delete = []
    for container, key in keys_to_delete:
        if isinstance(container, list):
            # If it's a list, we want to gather up the items to remove from the list.
            # We don't want to modify the list length yet, since removals will change all the other
            # list indexes. Instead, we'll get the actual items at those indexes to remove later.
            list_items_to_delete.append((container, container[key]))
        else:
            del container[key]

    # Finally, now that we have all the items we want to remove the lists, we'll remove those
    # items specifically from the lists.
    for containing_list, item in list_items_to_delete:
        containing_list.remove(item)


def normalise_path(path):
    """
    Converts a path to use correct path separator.
    Raises an PathConversionError if the path has a
    trailing slash.
    :param path: A directory path
    :type path: str
    :raises: sceptre.exceptions.PathConversionError
    :returns: A normalised path with forward slashes.
    :returns: string
    """
    if sep == "/":
        path = path.replace("\\", "/")
    elif sep == "\\":
        path = path.replace("/", "\\")
    if path.endswith("/") or path.endswith("\\"):
        raise PathConversionError(
            "'{0}' is an invalid path string. Paths should "
            "not have trailing slashes.".format(path)
        )
    return path


def sceptreise_path(path):
    """
    Converts a path to use correct sceptre path separator.
    Raises an PathConversionError if the path has a
    trailing slash.
    :param path: A directory path
    :type path: str
    :raises: sceptre.exceptions.PathConversionError
    :returns: A normalised path with forward slashes.
    :returns: string
    """
    path = path.replace("\\", "/")
    if path.endswith("/") or path.endswith("\\"):
        raise PathConversionError(
            "'{0}' is an invalid path string. Paths should "
            "not have trailing slashes.".format(path)
        )
    return path


@contextmanager
def null_context():
    """A context manager that does nothing. This is identical to the nullcontext in py3.7+, but isn't
    available in py3.6, so providing it here instead.
    """
    yield


def extract_datetime_from_aws_response_headers(
    boto_response: dict,
) -> Optional[datetime]:
    """Returns a datetime.datetime extracted from the response metadata in a
    boto response or None if it's unable to find or parse one.
    :param boto_response: A dictionary returned from a boto client call
    :returns a datetime.datetime or None
    """
    if boto_response is None:
        return None
    try:
        return dateutil.parser.parse(
            boto_response["ResponseMetadata"]["HTTPHeaders"]["date"]
        )
    except (KeyError, dateutil.parser.ParserError):
        # We expect a KeyError if the date isn't present in the response. We
        # expect a ParserError if it's not well-formed. Any other error we want
        # to pass along.
        return None


def gen_repr(instance: Any, class_label: str = None, attributes: List[str] = []) -> str:
    """
    Returns a standard __repr__ based on instance attributes.
    :param instance: The instance to represent (`self`).
    :param class_label: Override the name of the class found through introspection.
    :param attributes: List the attributes to include the in representation.
    :returns: A string representation of `instance`
    """
    if not class_label:
        class_label = instance.__class__.__name__
    attr_str = ", ".join(
        [f"{a}={repr(instance.__getattribute__(a))}" for a in attributes]
    )
    return f"{class_label}({attr_str})"


def create_deprecated_alias_property(
    alias_from: str, alias_to: str, deprecated_in: str, removed_in: Optional[str]
) -> property:
    """Creates a property object with a deprecated getter and a deprecated setter that emit warnings
    when used, aliasing to their renamed property names.

    :param alias_from: The name of the attribute that is deprecated and that needs to be aliased
    :param alias_to: The name of the attribute to alias the deprecated field to.
    :param deprecated_in: The version in which the property is deprecated.
    :param removed_in: The version when it will be removed, after which the alias will no longer work.
        This value can be None, indicating that removal is not yet planned.
    :return: A property object to be assigned directly onto a class.
    """

    def getter(self):
        return getattr(self, alias_to)

    getter.__name__ = alias_from

    def setter(self, value):
        setattr(self, alias_to, value)

    setter.__name__ = alias_from

    deprecation_kwargs = dict(
        deprecated_in=deprecated_in,
        removed_in=removed_in,
        current_version=__version__,
        details=(
            f'It is being renamed to "{alias_to}". You should migrate all uses of "{alias_from}" to '
            f"that in order to avoid future breakage."
        ),
    )

    deprecated_getter = deprecation.deprecated(**deprecation_kwargs)(getter)
    deprecated_setter = deprecation.deprecated(**deprecation_kwargs)(setter)

    deprecated_property = property(deprecated_getter, deprecated_setter)
    return deprecated_property



# -*- coding: utf-8 -*-
import pathlib
import tempfile
import sceptre.template_handlers.helper as helper

from sceptre.exceptions import UnsupportedTemplateFileTypeError
from sceptre.template_handlers import TemplateHandler


class S3(TemplateHandler):
    """
    Template handler that can resolve templates from S3.  Raw CFN templates
    with extension (.json, .yaml, .template) are deployed directly from memory
    while references to jinja (.j2) and python (.py) templates are downloaded,
    transformed into CFN templates then deployed to AWS.
    """

    def __init__(self, *args, **kwargs):
        super(S3, self).__init__(*args, **kwargs)

    def schema(self):
        return {
            "type": "object",
            "properties": {"path": {"type": "string"}},
            "required": ["path"],
        }

    def handle(self):
        """
        handle template in S3 bucket
        """
        input_path = self.arguments["path"]
        path = pathlib.Path(input_path)

        standard_template_suffix = [".json", ".yaml", ".template"]
        jinja_template_suffix = [".j2"]
        python_template_suffix = [".py"]
        supported_suffix = (
            standard_template_suffix + jinja_template_suffix + python_template_suffix
        )

        if path.suffix not in supported_suffix:
            raise UnsupportedTemplateFileTypeError(
                "Template has file extension %s. Only %s are supported.",
                path.suffix,
                ",".join(supported_suffix),
            )

        try:
            template = self._get_template(path)
            if path.suffix in jinja_template_suffix + python_template_suffix:
                file = tempfile.NamedTemporaryFile(prefix=path.stem)
                with file as f:
                    f.write(template)
                    f.seek(0)
                    f.read()
                    if path.suffix in jinja_template_suffix:
                        template = helper.render_jinja_template(
                            f.name,
                            {"sceptre_user_data": self.sceptre_user_data},
                            self.stack_group_config.get("j2_environment", {}),
                        )
                    elif path.suffix in python_template_suffix:
                        template = helper.call_sceptre_handler(
                            f.name, self.sceptre_user_data
                        )

        except Exception as e:
            helper.print_template_traceback(path)
            raise e

        return template

    def _get_template(self, path):
        """
        Get template from S3 bucket

        :param path: The path to the object in the bucket
        :type: str
        :returns: The body of the CloudFormation template.
        :rtype: str
        """
        self.logger.debug("Downloading file from S3: %s", path)
        bucket = path.parts[0]
        key = "/".join(path.parts[1:])

        try:
            response = self.connection_manager.call(
                service="s3",
                command="get_object",
                kwargs={"Bucket": bucket, "Key": key},
            )
            return response["Body"].read()
        except Exception as e:
            self.logger.critical(e)
            raise e



import click

from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions
from sceptre.cli.helpers import confirmation
from sceptre.cli.helpers import stack_status_exit_code
from sceptre.plan.plan import SceptrePlan

from colorama import Fore, Style


@click.command(name="delete", short_help="Deletes a stack or a change set.")
@click.argument("path")
@click.argument("change-set-name", required=False)
@click.option("-y", "--yes", is_flag=True, help="Assume yes to all questions.")
@click.pass_context
@catch_exceptions
def delete_command(ctx, path, change_set_name, yes):
    """
    Deletes a stack for a given config PATH. Or if CHANGE_SET_NAME is specified
    deletes a change set for stack in PATH.
    \f

    :param path: Path to execute command on.
    :type path: str
    :param change_set_name: The name of the change set to use - optional
    :type change_set_name: str
    :param yes: Flag to answer yes to all CLI questions.
    :type yes: bool
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
        full_scan=True,
    )

    plan = SceptrePlan(context)
    plan.resolve(command="delete", reverse=True)

    if change_set_name:
        delete_msg = (
            "The Change Set will be delete on the following stacks, if applicable:\n"
        )
    else:
        delete_msg = "The following stacks, in the following order, will be deleted:\n"

    dependencies = ""
    for stack in plan:
        dependencies += "{}{}{}\n".format(Fore.YELLOW, stack.name, Style.RESET_ALL)

    print(delete_msg + "{}".format(dependencies))

    confirmation(
        plan.delete.__name__, yes, change_set=change_set_name, command_path=path
    )
    if change_set_name:
        plan.delete_change_set(change_set_name)
    else:
        responses = plan.delete()
        exit(stack_status_exit_code(responses.values()))



import subprocess
from sceptre.hooks import Hook
from sceptre.exceptions import InvalidHookArgumentTypeError


class Cmd(Hook):
    """
    Cmd implements a Sceptre hook which can run arbitrary commands.
    """

    def __init__(self, *args, **kwargs):
        super(Cmd, self).__init__(*args, **kwargs)

    def run(self):
        """
        Executes a command through the shell.

        See hooks documentation for details.

        :raises: sceptre.exceptions.InvalidHookArgumentTypeError invalid input
        :raises: CalledProcessError failed command
        :raises: FileNotFoundError missing shell
        :raises: PermissionError non-executable shell
        """
        envs = self.stack.connection_manager.create_session_environment_variables()

        if isinstance(self.argument, str) and self.argument != "":
            command_to_run = self.argument
            shell = None

        elif (
            isinstance(self.argument, dict)
            and set(self.argument) == {"run", "shell"}
            and isinstance(self.argument["run"], str)
            and isinstance(self.argument["shell"], str)
            and self.argument["run"] != ""
            and self.argument["shell"] != ""
        ):
            command_to_run = self.argument["run"]
            shell = self.argument["shell"]

        else:
            raise InvalidHookArgumentTypeError(
                "A cmd hook requires either a string argument or an object with "
                "`run` and `shell` keys with string values. "
                f"You gave `{self.argument!r}`."
            )

        subprocess.check_call(command_to_run, shell=True, env=envs, executable=shell)



import logging
import webbrowser
import click

from sceptre.cli.dump import dump_template
from sceptre.cli.helpers import catch_exceptions, write
from sceptre.context import SceptreContext
from sceptre.helpers import null_context
from sceptre.plan.plan import SceptrePlan
from sceptre.resolvers.placeholders import use_resolver_placeholders_on_error

logger = logging.getLogger(__name__)


@click.command(name="validate", short_help="Validates the template.")
@click.option(
    "-n",
    "--no-placeholders",
    is_flag=True,
    help="If True, no placeholder values will be supplied for resolvers that cannot be resolved.",
)
@click.argument("path")
@click.pass_context
@catch_exceptions
def validate_command(ctx, no_placeholders, path):
    """
    Validates the template used for stack in PATH.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)

    execution_context = (
        null_context() if no_placeholders else use_resolver_placeholders_on_error()
    )
    with execution_context:
        responses = plan.validate()

    for stack, response in responses.items():
        if response["ResponseMetadata"]["HTTPStatusCode"] == 200:
            del response["ResponseMetadata"]
            click.echo("Template {} is valid. Template details:\n".format(stack.name))
        write(response, context.output_format)


@click.command(name="generate", short_help="Prints the template.")
@click.option(
    "-n",
    "--no-placeholders",
    is_flag=True,
    help="If True, no placeholder values will be supplied for resolvers that cannot be resolved.",
)
@click.argument("path")
@click.pass_context
@catch_exceptions
def generate_command(ctx: click.Context, no_placeholders: bool, path: str):
    """
    Prints the template used for stack in PATH.

    This command is aliased to the dump template command for legacy support reasons. It's the same
    as running `sceptre dump template`.

    \f
    :param no_placeholders: If True, will disable placeholders for unresolvable resolvers. By
        default, placeholders will be active.
    :param path: Path to execute the command on.
    """
    ctx.forward(dump_template)


@click.command(name="estimate-cost", short_help="Estimates the cost of the template.")
@click.argument("path")
@click.pass_context
@catch_exceptions
def estimate_cost_command(ctx, path):
    """
    Prints a URI to STOUT that provides an estimated cost based on the
    resources in the stack. This command will also attempt to open a web
    browser with the returned URI.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)
    responses = plan.estimate_cost()

    for stack, response in responses.items():
        if response["ResponseMetadata"]["HTTPStatusCode"] == 200:
            del response["ResponseMetadata"]
            click.echo("View the estimated cost for {} at:".format(stack.name))
            response = response["Url"]
            webbrowser.open(response, new=2)
        write(response + "\n", "text")


@click.command(name="fetch-remote-template", short_help="Prints the remote template.")
@click.argument("path")
@click.pass_context
@catch_exceptions
def fetch_remote_template_command(ctx, path):
    """
    Prints the remote template used for stack in PATH.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)
    responses = plan.fetch_remote_template()
    output = []
    for stack, template in responses.items():
        if template is None:
            logger.warning(f"{stack.external_name} does not exist")
        else:
            output.append(template)

    write(output, context.output_format)



# -*- coding: utf-8 -*-

import logging
import warnings
import sys

from importlib.metadata import version

__author__ = "SceptreOrg"
__email__ = "sceptreorg@gmail.com"
__version__ = version(__package__ or __name__)


# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
class NullHandler(logging.Handler):  # pragma: no cover
    def emit(self, record):
        pass


if not sys.warnoptions:
    warnings.filterwarnings("default", category=DeprecationWarning, module="sceptre")

logging.getLogger("sceptre").addHandler(NullHandler())



from sceptre.resolvers import Resolver


class NoValue(Resolver):
    """This resolver resolves to nothing, functioning just like the AWS::NoValue special value. When
    assigned to a resolvable Stack property, it will remove the config key/value from the stack or
    the container on the stack where it has been assigned, as if this value wasn't assigned at all.

    This is mostly useful for simplifying conditional logic on Stack and StackGroup config files
    where, if a certain condition is met, a value is passed, otherwise it's not passed at all.
    """

    def resolve(self) -> None:
        return None



# -*- coding: utf-8 -*-

import functools
import logging
import shlex

from botocore.exceptions import ClientError

from sceptre.exceptions import (
    DependencyStackMissingOutputError,
    StackDoesNotExistError,
    SceptreException,
)

from sceptre.helpers import normalise_path, sceptreise_path
from sceptre.resolvers import Resolver

TEMPLATE_EXTENSION = ".yaml"


class StackOutputBase(Resolver):
    """
    A abstract base class which provides methods for getting Stack outputs.
    """

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(__name__)
        super(StackOutputBase, self).__init__(*args, **kwargs)

    def _get_output_value(
        self, stack_name, output_key, profile=None, region=None, sceptre_role=None
    ):
        """
        Attempts to get the Stack output named by ``output_key``

        :param stack_name: Name of the Stack to collect output for.
        :type stack_name: str
        :param output_key: The name of the Stack output in which to return.
        :type output_key: str
        :returns: Stack output value.
        :rtype: str
        :raises: sceptre.exceptions.DependencyStackMissingOutputError
        """
        outputs = self._get_stack_outputs(stack_name, profile, region, sceptre_role)

        try:
            return outputs[output_key]
        except KeyError:
            raise DependencyStackMissingOutputError(
                "The Stack '{0}' does not have an output named '{1}'".format(
                    stack_name, output_key
                )
            )

    @functools.lru_cache(maxsize=4096)
    def _get_stack_outputs(
        self, stack_name, profile=None, region=None, sceptre_role=None
    ):
        """
        Communicates with AWS CloudFormation to fetch outputs from a specific
        Stack.

        :param stack_name: Name of the Stack to collect output for.
        :type stack_name: str
        :returns: A formatted version of the Stack outputs.
        :rtype: dict
        :raises: sceptre.stack.DependencyStackNotLaunchedException
        """
        self.logger.debug("Collecting outputs from '{0}'...".format(stack_name))
        connection_manager = self.stack.connection_manager

        try:
            response = connection_manager.call(
                service="cloudformation",
                command="describe_stacks",
                kwargs={"StackName": stack_name},
                profile=profile,
                region=region,
                stack_name=stack_name,
                sceptre_role=sceptre_role,
            )
        except ClientError as e:
            if "does not exist" in e.response["Error"]["Message"]:
                raise StackDoesNotExistError(e.response["Error"]["Message"])
            else:
                raise e
        else:
            outputs = response["Stacks"][0].get("Outputs", {})

        self.logger.debug("Outputs: {0}".format(outputs))

        formatted_outputs = dict(
            (output["OutputKey"], output["OutputValue"]) for output in outputs
        )

        return formatted_outputs


class StackOutput(StackOutputBase):
    """
    Resolver for retrieving the value of a Stack output within the current
    Sceptre StackGroup. Adds the target Stack to the dependencies of the
    Stack using the Resolver.

    :param argument: The Stack name and output name to get.
    :type argument: str in the format ``"<stack name>::<output key>"``
    """

    def __init__(self, *args, **kwargs):
        super(StackOutput, self).__init__(*args, **kwargs)

    def setup(self):
        """
        Adds dependency to a Stack.
        """
        try:
            dep_stack_name, self.output_key = self.argument.split("::")
        except ValueError as err:
            raise SceptreException(
                "!stack_output arg should match STACK_NAME::OUTPUT_KEY"
            ) from err

        self.dependency_stack_name = sceptreise_path(normalise_path(dep_stack_name))
        self.stack.dependencies.append(self.dependency_stack_name)

    def resolve(self):
        """
        Retrieves the value of an output of an internal Stack.

        :returns: The value of the Stack output.
        :rtype: str
        """
        self.logger.debug("Resolving Stack output: {0}".format(self.argument))
        friendly_stack_name = self.dependency_stack_name.replace(TEMPLATE_EXTENSION, "")

        stack = next(
            stack
            for stack in self.stack.dependencies
            if stack.name == friendly_stack_name
        )

        stack_name = "-".join(
            [stack.project_code, friendly_stack_name.replace("/", "-")]
        )

        return self._get_output_value(
            stack_name,
            self.output_key,
            profile=stack.profile,
            region=stack.region,
            sceptre_role=stack.sceptre_role,
        )


class StackOutputExternal(StackOutputBase):
    """
    Resolver for retrieving the value of an output of any Stack within the
    current Sceptre stack_group's account and region.

    :param argument: The Stack name and output name to get.
    :type argument: str in the format ``"<full stack name>::<output key>"``
    """

    def __init__(self, *args, **kwargs):
        super(StackOutputExternal, self).__init__(*args, **kwargs)

    def resolve(self):
        """
        Retrieves the value of CloudFormation output of the external Stack

        :returns: The value of the Stack output.
        :rtype: str
        """
        self.logger.debug("Resolving external Stack output: {0}".format(self.argument))

        arguments = shlex.split(self.argument)

        if not arguments:
            message = "!stack_output_external requires at least one argument"
            raise SceptreException(message)

        stack_argument = arguments[0]
        stack_args = iter(stack_argument.split("::"))

        try:
            dependency_stack_name = next(stack_args)
            output_key = next(stack_args)

        except StopIteration as err:
            message = "!stack_output_external arg should match STACK_NAME::OUTPUT_KEY"
            raise SceptreException(message) from err

        profile = region = sceptre_role = None

        if len(arguments) > 1:
            extra_args = iter(arguments[1].split("::"))

            profile = next(extra_args, None)
            region = next(extra_args, None)
            sceptre_role = next(extra_args, None)

            try:
                next(extra_args)
                message = (
                    "!stack_output_external second arg should be "
                    "in the format 'PROFILE[::REGION[::SCEPTRE_ROLE]]'"
                )
                raise SceptreException(message)

            except StopIteration:
                pass

        return self._get_output_value(
            dependency_stack_name, output_key, profile, region, sceptre_role
        )



# -*- coding: utf-8 -*-
import abc
import logging
from contextlib import contextmanager
from threading import RLock
from typing import Any, TYPE_CHECKING, Type, Union, TypeVar

from sceptre.exceptions import InvalidResolverArgumentError
from sceptre.helpers import _call_func_on_values, delete_keys_from_containers
from sceptre.logging import StackLoggerAdapter
from sceptre.resolvers.placeholders import (
    create_placeholder_value,
    are_placeholders_enabled,
    PlaceholderType,
)

if TYPE_CHECKING:
    from sceptre import stack

T_Container = TypeVar("T_Container", bound=Union[dict, list])
Self = TypeVar("Self")


class RecursiveResolve(Exception):
    pass


class CustomYamlTagBase:
    """A base class for custom Yaml Elements (i.e. hooks and resolvers).

    This base class takes care of common functionality needed by subclasses:
    | * logging setup (associated with stacks)
    | * Creating unique clones associated with individual stacks (applied recursively down to all
    |   resolvers in the argument)
    | * On-stack-connect setup that might be needed once connected to a Stack (applied recursively down to
    |   all resolvers in the argument)
    | * Automatically resolving resolvers in the argument when accessing self.argument

    """

    logger = logging.getLogger(__name__)

    def __init__(self, argument: Any = None, stack: "stack.Stack" = None):
        """Initializes a custom yaml tag object.

        :param argument: The argument passed to the yaml tag. This could be a string, number, list,
            or dict. Resolvers are supported in the argument, but they must be in either list or dict
            arguments.
        :param stack: The stack object associated with this instance. NOTE: When first instantiated
            from loading the YAML, there will be no Stack. A Stack instance will only be passed into
            the instance when "clone_for_stack" is invoked when the tag is associated with a specific
            stack.
        """
        if stack is not None:
            self.logger = StackLoggerAdapter(self.logger, stack.name)

        self.stack = stack

        self._argument = argument
        self._argument_is_resolved = False

    @property
    def argument(self) -> Any:
        """This is the resolver or hook's argument.

        This property will resolve all nested resolvers inside the argument, but only if this
        instance has been associated with a Stack.

        Resolving nested resolvers will result in their values being replaced in the dict/list they
        were in with their resolved value, so we won't have to resolve them again.

        Any resolvers that "resolve to nothing" (i.e. return None) will be removed from the dict/list
        they were in.

        If this property is accessed BEFORE the instance has a stack, it will return
        the raw argument value. This is to safeguard any __init__() behaviors from triggering
        resolution prematurely.
        """
        if self.stack is not None and not self._argument_is_resolved:
            # Since resolving the argument updates the argument list or dict so that there aren't
            # resolvers in it any more, we only need to do this resolution once per instance.
            self._resolve_argument()

        return self._argument

    @argument.setter
    def argument(self, value):
        self._argument = value

    def _resolve_argument(self):
        """Resolves all argument resolvers recursively."""

        keys_to_delete = []

        def resolve(containing_list_or_dict, key, obj: Resolver):
            result = obj.resolve()
            # If the resolver "resolves to nothing", then it should get deleted out of its container.
            if result is None:
                keys_to_delete.append((containing_list_or_dict, key))
            else:
                containing_list_or_dict[key] = result

        _call_func_on_values(resolve, self._argument, Resolver)
        delete_keys_from_containers(keys_to_delete)

        self._argument_is_resolved = True

    def _recursively_setup(self):
        """Ensures all nested resolvers in this resolver's argument are also setup when this
        instance's setup method is called.
        """
        self.setup()

        def setup_nested(containing_list_or_dict, key, obj: Resolver):
            obj._recursively_setup()

        _call_func_on_values(setup_nested, self._argument, Resolver)

    def _recursively_clone(self: Self, stack: "stack.Stack") -> Self:
        """Recursively clones the instance and its arguments.

        The returned instance will have an identical argument that is a different memory reference,
        so that instances inherited from a stack group and applied across multiple stacks are
        independent of each other.

        Furthermore, all nested resolvers in this resolver's argument will also be cloned to ensure
        they themselves are also independent and fully configured for the current stack.
        """

        def recursively_clone_arguments(obj):
            if isinstance(obj, Resolver):
                return obj._recursively_clone(stack)
            if isinstance(obj, list):
                return [recursively_clone_arguments(item) for item in obj]
            elif isinstance(obj, dict):
                return {
                    key: recursively_clone_arguments(val) for key, val in obj.items()
                }
            return obj

        argument = recursively_clone_arguments(self._argument)
        clone = type(self)(argument, stack)
        return clone

    def clone_for_stack(self: Self, stack: "stack.Stack") -> Self:
        """
        Obtains a clone of the current object, setup and ready for use for a given Stack instance.
        """
        clone = self._recursively_clone(stack)
        clone._recursively_setup()
        return clone

    def setup(self):
        """
        This method is called when the object is connected to a Stack instance.
        Implementation of this method in subclasses can be used to do any
        initial setup of the object.
        """
        pass  # pragma: no cover

    def __repr__(self) -> str:
        """Returns a string representation of the resolver.

        This is mostly used for resolver placeholders. In cases where we cannot resolve a resolver
        YET, such as when we're generating a template or diff and there's a dependency on a stack
        output of a stack that hasn't been deployed yet, placeholders need to render the resolver
        in a way that is useful.

        We use self._argument instead of self.argument because if there are resolvers nested in this
        resolver's argument and one of those cannot be resolved, we'll need those resolvers to also
        be converted to useful placeholder values.
        """
        as_str = f"!{self.__class__.__name__}"
        if self._argument is not None:
            as_str += f"({self._argument})"

        return as_str


class Resolver(CustomYamlTagBase, metaclass=abc.ABCMeta):
    """
    Resolver is an abstract base class that should be subclassed by all Resolvers.
    """

    @abc.abstractmethod
    def resolve(self):
        """
        An abstract method which must be overwritten by all inheriting classes.
        This method is called to retrieve the final desired value.
        Implementation of this method in subclasses must return a suitable
        object or primitive type.
        """
        pass  # pragma: no cover

    def raise_invalid_argument_error(self, message, from_: Exception = None):
        error_message = f"{self.stack.name} - {message}"
        if from_:
            raise InvalidResolverArgumentError(error_message) from from_
        raise InvalidResolverArgumentError(error_message)


class ResolvableProperty(abc.ABC):
    """
    This is an abstract base class for a descriptor used to store an attribute that have values
    associated with Resolver objects.

    :param name: Attribute suffix used to store the property in the instance.
    :param placeholder_type: The type of placeholder that should be returned, when placeholders are
        allowed, when a resolver can't be resolved.
    """

    def __init__(self, name: str, placeholder_type=PlaceholderType.explicit):
        self.name = "_" + name
        self.logger = logging.getLogger(__name__)
        self.placeholder_type = placeholder_type

        self._lock = RLock()

    def __get__(self, stack: "stack.Stack", stack_class: Type["stack.Stack"]) -> Any:
        """
        Attribute getter which resolves the resolver(s).

        :param stack: The Stack instance the property is being retrieved for
        :param stack_class: The class of the stack that the property is being retrieved for.
        :return: The attribute stored with the suffix ``name`` in the instance.
        :rtype: The obtained value, as resolved by the property
        """
        with self._lock, self._no_recursive_get(stack):
            if hasattr(stack, self.name):
                return self.get_resolved_value(stack, stack_class)

    def __set__(self, stack: "stack.Stack", value: Any):
        """
        Attribute setter which adds a stack reference to any resolvers in the
        data structure `value` and calls the setup method.

        :param stack: The Stack instance the value is being set onto
        :param value: The value being set on the property
        """
        with self._lock:
            self.assign_value_to_stack(stack, value)

    @contextmanager
    def _no_recursive_get(self, stack: "stack.Stack"):
        # We don't care about recursive gets on the same property but different Stack instances,
        # only recursive gets on the same stack. Some Resolvers access the same property on OTHER
        # stacks and that actually shouldn't be a problem. Remember, these descriptor instances are
        # set on the CLASS and so instance variables on them are shared across all classes that
        # access them. Thus, we set this "get_in_progress" attribute on the stack instance rather
        # than the descriptor instance.
        get_status_name = f"_{self.name}_get_in_progress"
        if getattr(stack, get_status_name, False):
            raise RecursiveResolve(
                f"Resolving Stack.{self.name[1:]} required resolving itself"
            )
        setattr(stack, get_status_name, True)
        try:
            yield
        finally:
            setattr(stack, get_status_name, False)

    @abc.abstractmethod
    def get_resolved_value(
        self, stack: "stack.Stack", stack_class: Type["stack.Stack"]
    ) -> Any:
        """Implement this method to return the value of the resolvable_property."""
        pass

    @abc.abstractmethod
    def assign_value_to_stack(self, stack: "stack.Stack", value: Any):
        """Implement this method to assign the value to the resolvable property."""
        pass

    def resolve_resolver_value(self, resolver: "Resolver") -> Any:
        """Returns the resolved parameter value.

        If the resolver happens to raise an error and placeholders are currently allowed for resolvers,
        a placeholder will be returned instead of reraising the error.

        :param resolver: The resolver to resolve.
        :return: The resolved value (or placeholder, in certain circumstances)
        """
        try:
            return resolver.resolve()
        except RecursiveResolve:
            # Recursive resolve issues shouldn't be masked by a placeholder.
            raise
        except Exception:
            if are_placeholders_enabled():
                placeholder_value = create_placeholder_value(
                    resolver, self.placeholder_type
                )

                self.logger.debug(
                    "Error encountered while resolving the resolver. This is allowed for the current "
                    f"operation. Resolving it to a placeholder value instead: {placeholder_value}"
                )
                return placeholder_value
            raise

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.name[1:]})>"


class ResolvableContainerProperty(ResolvableProperty):
    """
    This is a descriptor class used to store an attribute that may CONTAIN
    Resolver objects. Resolvers will be resolved upon access of this property.
    When resolvers are resolved, they will be replaced in the container with their
    resolved value, in order to avoid redundant resolutions.

    Supports nested dictionary and lists.

    :param name: Attribute suffix used to store the property in the instance.
    :type name: str
    """

    def __get__(
        self, stack: "stack.Stack", stack_class: Type["stack.Stack"]
    ) -> T_Container:
        container = super().__get__(stack, stack_class)

        with self._lock:
            # Resolve any deferred resolvers, now that the recursive get lock has been released.
            self._resolve_deferred_resolvers(stack, container)

        return container

    def get_resolved_value(
        self, stack: "stack.Stack", stack_class: Type["stack.Stack"]
    ) -> T_Container:
        """Obtains the resolved value for this property. Any resolvers that resolve to None will have
        their key/index removed from their dict/list where they are. Other resolvers will have their
        key/index's value replace with the resolved value to avoid redundant resolutions.

        :param stack: The Stack instance to obtain the value for
        :param stack_class: The class of the Stack instance.
        :return: The fully resolved container.
        """
        keys_to_delete = []

        def resolve(attr: Union[dict, list], key: Union[int, str], value: Resolver):
            # Update the container key's value with the resolved value, if possible...
            try:
                result = self.resolve_resolver_value(value)
                if result is None:
                    self.logger.debug(
                        f"Removing item {key} because resolver returned None."
                    )
                    # We gather up resolvers (and their immediate containers) that resolve to None,
                    # since that really means the resolver resolves to nothing. This is not common,
                    # but should be supported. We gather these rather than immediately remove them
                    # because this function is called in the context of looping over that attr, so
                    # we cannot alter its size until after the loop is complete.
                    keys_to_delete.append((attr, key))
                else:
                    attr[key] = result
            except RecursiveResolve:
                # It's possible that resolving the resolver might attempt to access another
                # resolvable property's value in this same container. In this case, we'll delay
                # resolution and instead return a ResolveLater so the value can be resolved outside
                # this recursion.
                attr[key] = self.ResolveLater(
                    stack,
                    self.name,
                    key,
                    lambda: value.resolve(),
                )

        container = getattr(stack, self.name)
        _call_func_on_values(resolve, container, Resolver)
        delete_keys_from_containers(keys_to_delete)

        return container

    def assign_value_to_stack(self, stack: "stack.Stack", value: Union[dict, list]):
        """Assigns a COPY of the specified value to the stack instance. This method copies the value
        rather than directly assigns it to avoid bugs related to shared objects in memory.

        :param stack: The stack to assign the value to
        :param value: The value to assign
        """
        cloned = self._clone_container_with_resolvers(value, stack)
        setattr(stack, self.name, cloned)

    def _clone_container_with_resolvers(
        self, container: T_Container, stack: "stack.Stack"
    ) -> T_Container:
        """Recurses into the container, cloning and setting up resolvers and creating a copy of all
        nested containers.

        :param container: The container being recursed into and cloned
        :param stack: The stack the container is being copied for
        :return: The fully copied container with resolvers fully set up.
        """

        def recurse(obj):
            if isinstance(obj, Resolver):
                return obj.clone_for_stack(stack)
            if isinstance(obj, list):
                return [recurse(item) for item in obj]
            elif isinstance(obj, dict):
                return {key: recurse(val) for key, val in obj.items()}
            return obj

        return recurse(container)

    def _resolve_deferred_resolvers(self, stack: "stack.Stack", container: T_Container):
        def raise_if_not_resolved(attr, key, value):
            # If this function has been hit, it means that after attempting to resolve all the
            # ResolveLaters, there STILL are ResolveLaters left in the container. Rather than
            # continuing to try to resolve (possibly infinitely), we'll raise a RecursiveGet to
            # break that infinite loop. This situation would happen if a resolver accesses a resolver
            # in the same container, which then accesses another resolver (possibly the same one) in
            # the same container.
            raise RecursiveResolve(
                f"Resolving Stack.{self.name[1:]} required resolving itself"
            )

        has_been_resolved_attr_name = f"{self.name}_is_resolved"
        if not getattr(stack, has_been_resolved_attr_name, False):
            # We set it first rather than after to avoid entering this block again on this property
            # for this stack.
            setattr(stack, has_been_resolved_attr_name, True)
            _call_func_on_values(
                lambda attr, key, value: value(), container, self.ResolveLater
            )
            # Search the container to see if there are any ResolveLaters left;
            # Raise a RecursiveResolve if there are.
            _call_func_on_values(raise_if_not_resolved, container, self.ResolveLater)

    class ResolveLater:
        """Represents a value that could not yet be resolved but can be resolved in the future."""

        def __init__(self, instance, name, key, resolution_function):
            self._instance = instance
            self._name = name
            self._key = key
            self._resolution_function = resolution_function

        def __call__(self):
            """Resolve the value."""
            attr = getattr(self._instance, self._name)
            result = self._resolution_function()
            if result is None:
                self.logger.debug(
                    f"Removing item {self._key} because resolver returned None."
                )
                del attr[self._key]
            else:
                attr[self._key] = result


class ResolvableValueProperty(ResolvableProperty):
    """
    This is a descriptor class used to store an attribute that may BE a single
    Resolver object. If it is a resolver, it will be resolved upon access of this property.
    When resolved, the resolved value will replace the resolver on the stack in order to avoid
    redundant resolutions.

    :param name: Attribute suffix used to store the property in the instance.
    :type name: str
    """

    def get_resolved_value(
        self, stack: "stack.Stack", stack_class: Type["stack.Stack"]
    ) -> Any:
        """Gets the fully-resolved value from the property. Resolvers will be replaced on the stack
        instance with their resolved value to avoid redundant resolutions.

        :param stack: The Stack instance to obtain the value from
        :param stack_class: The class of the Stack instance
        :return: The fully resolved value
        """
        raw_value = getattr(stack, self.name)
        if isinstance(raw_value, Resolver):
            value = self.resolve_resolver_value(raw_value)
            # Overwrite the stored resolver value with the resolved value to avoid resolving the
            # same value multiple times.
            setattr(stack, self.name, value)
        else:
            value = raw_value

        return value

    def assign_value_to_stack(self, stack: "stack.Stack", value: Any):
        """Assigns the value to the Stack instance passed, setting up and cloning the value if it
        is a Resolver.

        :param stack: The Stack instance to set the value on
        :param value: The value to set
        """
        if isinstance(value, Resolver):
            value = value.clone_for_stack(stack)
        setattr(stack, self.name, value)



import logging
import click

from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions, write
from sceptre.plan.plan import SceptrePlan

from typing import List, Dict

logger = logging.getLogger(__name__)


@click.group(name="list")
def list_group():
    """
    Commands for listing attributes of stacks.
    """
    pass


@list_group.command(name="resources")
@click.argument("path")
@click.pass_context
@catch_exceptions
def list_resources(ctx, path):
    """
    List resources for stack or stack_group.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )
    plan = SceptrePlan(context)

    responses = [
        response for response in plan.describe_resources().values() if response
    ]

    write(responses, context.output_format)


# flake8: noqa: C901
def write_outputs(export, responses, plan, context):
    """
    Helper function for list outputs.
    """
    # Legacy. This option was added in the initial commit of the project,
    # although its intended use case is unclear. It may relate to a feature
    # that had been removed prior to the initial commit.
    if export == "envvar":
        for response in responses:
            for stack in response.values():
                for output in stack:
                    write(
                        "export SCEPTRE_{0}='{1}'".format(
                            output.get("OutputKey"), output.get("OutputValue")
                        ),
                        "text",
                    )

    # Format outputs as !stack_output references.
    elif export == "stackoutput":
        for response in responses:
            for stack_name, stack in response.items():
                for output in stack:
                    write(
                        "!stack_output {0}.yaml::{1} [{2}]".format(
                            stack_name,
                            output.get("OutputKey"),
                            output.get("OutputValue"),
                        ),
                        "text",
                    )

    # Format outputs as !stack_output_external references.
    elif export == "stackoutputexternal":
        stack_names = {stack.name: stack.external_name for stack in plan.graph}
        for response in responses:
            for stack_name, stack in response.items():
                for output in stack:
                    write(
                        "!stack_output_external {0}::{1} [{2}]".format(
                            stack_names[stack_name],
                            output.get("OutputKey"),
                            output.get("OutputValue"),
                        ),
                        "text",
                    )

    # Legacy. The output here is somewhat confusing in that
    # outputs are organised in keys that only have meaning inside
    # Sceptre.
    else:
        write(responses, context.output_format)


@list_group.command(name="outputs")
@click.argument("path")
@click.option(
    "-e",
    "--export",
    type=click.Choice(["envvar", "stackoutput", "stackoutputexternal"]),
    help="Specify the export formatting.",
)
@click.pass_context
@catch_exceptions
def list_outputs(ctx, path, export):
    """
    List outputs for stack.
    \f
    :param path: Path to execute the command on.
    :type path: str
    :param export: Specify the export formatting.
    :type export: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path", None),
        user_variables=ctx.obj.get("user_variables", {}),
        options=ctx.obj.get("options", {}),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)
    responses = [response for response in plan.describe_outputs().values() if response]

    write_outputs(export, responses, plan, context)


@list_group.command(name="change-sets")
@click.option("-U", "--url", is_flag=True, help="Instead write a URL.")
@click.argument("path")
@click.pass_context
@catch_exceptions
def list_change_sets(ctx, path, url):
    """
    List change sets for stack.
    \f

    :param path: Path to execute the command on.
    :type path: str
    :param url: Write out a console URL instead.
    :type url: bool
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        output_format=ctx.obj.get("output_format"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)

    responses = [
        response for response in plan.list_change_sets(url).values() if response
    ]

    for response in responses:
        write(response, context.output_format)


@list_group.command(name="stacks")
@click.argument("path")
@click.pass_context
@catch_exceptions
def list_stacks(ctx, path):
    """
    List sceptre stack config attributes,
    \f

    :param path: Path to execute the command on or path to stack group
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        output_format=ctx.obj.get("output_format"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )

    plan = SceptrePlan(context)

    output = {f"{stack.name}.yaml": stack.external_name for stack in plan.graph}
    output_format = "json" if context.output_format == "json" else "yaml"
    write(output, output_format)



import abc
import logging
from functools import wraps
from typing import TYPE_CHECKING, List

from sceptre.helpers import _call_func_on_values
from sceptre.resolvers import CustomYamlTagBase

if TYPE_CHECKING:
    from sceptre.stack import Stack


class Hook(CustomYamlTagBase, metaclass=abc.ABCMeta):
    """
    Hook is an abstract base class that should be subclassed by all hooks.
    """

    logger = logging.getLogger(__name__)

    @abc.abstractmethod
    def run(self):
        """
        run is an abstract method which must be overwritten by all
        inheriting classes. Run should execute the logic of the hook.
        """
        pass  # pragma: no cover


class HookProperty(object):
    """
    This is a descriptor class used to store an attribute that may contain
    Hook objects. Used to setup Hooks when added as a attribute. Supports
    nested dictionary and lists.

    :param name: Attribute suffix used to store the property in the instance.
    :type name: str
    """

    def __init__(self, name):
        self.name = "_" + name
        self.logger = logging.getLogger(__name__)

    def __get__(self, instance, type):
        """
        Attribute getter for Hook containing data structure.

        :return: The attribute stored with the suffix ``name`` in the instance.
        :rtype: dict or list
        """
        return getattr(instance, self.name)

    def __set__(self, instance: "Stack", value):
        """
        Attribute setter which adds a stack reference to any hooks in the
        data structure `value` and calls the setup method.

        """

        def setup(attr, key, value: Hook):
            attr[key] = clone = value.clone_for_stack(instance)
            clone.setup()

        _call_func_on_values(setup, value, Hook)
        setattr(instance, self.name, value)


def execute_hooks(hooks):
    """
    Searches through dictionary or list for Resolver objects and replaces
    them with the resolved value. Supports nested dictionaries and lists.
    Does not detect Resolver objects used as keys in dictionaries.

    :param attr: A complex data structure to search through.
    :type attr: dict or list
    :return: A complex data structure without Resolver objects.
    :rtype: dict or list
    """
    if isinstance(hooks, list):
        for hook in hooks:
            if isinstance(hook, Hook):
                hook.run()


def add_stack_hooks(func):
    """
    A function decorator to trigger the before and after hooks, relative
    to the decorated function's name.
    :param func: a function that operates on a stack
    :type func: function
    """

    @wraps(func)
    def decorated(self, *args, **kwargs):
        execute_hooks(self.stack.hooks.get("before_" + func.__name__))
        response = func(self, *args, **kwargs)
        execute_hooks(self.stack.hooks.get("after_" + func.__name__))

        return response

    return decorated


def add_stack_hooks_with_aliases(function_aliases: List[str]):
    """
    Returns a decorator to trigger the before and after hooks, relative to the decorated function's
    name AS WELL AS the passed function alias names.
    :param function_aliases: The list of OTHER functions to trigger hooks around.
    :return: The hook-triggering decorator.
    """

    def decorator(func):
        all_hook_names = [func.__name__] + function_aliases

        @wraps(func)
        def decorated(self, *args, **kwargs):
            for hook in all_hook_names:
                before_hook_name = f"before_{hook}"
                execute_hooks(self.stack.hooks.get(before_hook_name))

            response = func(self, *args, **kwargs)

            for hook in all_hook_names:
                after_hook_name = f"after_{hook}"
                execute_hooks(self.stack.hooks.get(after_hook_name))

            return response

        return decorated

    return decorator



from sceptre.resolvers import Resolver


class Join(Resolver):
    """This resolver allows you to join multiple strings together to form a single string. This is
    great for combining the outputs of multiple resolvers. This resolver works just like
    CloudFormation's ``!Join`` intrinsic function.

    The argument for this resolver should be a list with two elements: (1) A string to join the
    elements on and (2) a list of items to join.

    Example:

       parameters:
         BaseUrl: !join
           - ":"
           - - !stack_output my/app/stack.yaml::HostName
             - !stack_output my/other/stack.yaml::Port

    """

    def resolve(self):
        error_message = (
            "The argument to !join must be a 2-element list, where the first element is the join "
            "string and the second is a list of items to join."
        )
        if not isinstance(self.argument, list) or len(self.argument) != 2:
            self.raise_invalid_argument_error(error_message)

        delimiter, items_list = self.argument
        if not isinstance(delimiter, str) or not isinstance(items_list, list):
            self.raise_invalid_argument_error(error_message)

        string_items = map(str, items_list)
        joined = delimiter.join(string_items)
        return joined



# -*- coding: utf-8 -*-

"""
sceptre.config

This module implements a ConfigReader class, which is responsible for reading
and constructing Stacks.
"""

import collections
import copy
import datetime
import fnmatch
import logging
import sys
import yaml
import json

from os import environ, path, walk
from typing import Set, Tuple
from pathlib import Path
from jinja2 import Environment
from jinja2 import StrictUndefined
from jinja2 import FileSystemLoader
from jinja2 import select_autoescape
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from sceptre import __version__
from sceptre.exceptions import SceptreException
from sceptre.exceptions import DependencyDoesNotExistError
from sceptre.exceptions import InvalidConfigFileError
from sceptre.exceptions import InvalidSceptreDirectoryError
from sceptre.exceptions import VersionIncompatibleError
from sceptre.exceptions import ConfigFileNotFoundError
from sceptre.helpers import sceptreise_path, logging_level, write_debug_file
from sceptre.stack import Stack
from sceptre.config import strategies

ConfigAttributes = collections.namedtuple("Attributes", "required optional")


CONFIG_MERGE_STRATEGY_OVERRIDES = {
    "dependencies": strategies.LIST_STRATEGIES,
    "hooks": strategies.LIST_STRATEGIES,
    "notifications": strategies.LIST_STRATEGIES,
    "parameters": strategies.DICT_STRATEGIES,
    "sceptre_user_data": strategies.DICT_STRATEGIES,
    "stack_tags": strategies.DICT_STRATEGIES,
}

CONFIG_MERGE_STRATEGIES = {
    "dependencies": strategies.list_join,
    "dependencies_inheritance": strategies.child_or_parent,
    "hooks": strategies.child_wins,
    "hooks_inheritance": strategies.child_or_parent,
    "iam_role": strategies.child_wins,
    "sceptre_role": strategies.child_wins,
    "iam_role_session_duration": strategies.child_wins,
    "sceptre_role_session_duration": strategies.child_wins,
    "notifications": strategies.child_wins,
    "notifications_inheritance": strategies.child_or_parent,
    "on_failure": strategies.child_wins,
    "parameters": strategies.child_wins,
    "parameters_inheritance": strategies.child_or_parent,
    "profile": strategies.child_wins,
    "project_code": strategies.child_wins,
    "protect": strategies.child_wins,
    "region": strategies.child_wins,
    "required_version": strategies.child_wins,
    "role_arn": strategies.child_wins,
    "cloudformation_service_role": strategies.child_wins,
    "sceptre_user_data": strategies.child_wins,
    "sceptre_user_data_inheritance": strategies.child_or_parent,
    "stack_name": strategies.child_wins,
    "stack_tags": strategies.child_wins,
    "stack_tags_inheritance": strategies.child_or_parent,
    "stack_timeout": strategies.child_wins,
    "template_bucket_name": strategies.child_wins,
    "template_key_value": strategies.child_wins,
    "template": strategies.child_wins,
    "template_path": strategies.child_wins,
    "ignore": strategies.child_wins,
    "obsolete": strategies.child_wins,
}


STACK_GROUP_CONFIG_ATTRIBUTES = ConfigAttributes(
    {"project_code", "region"},
    {
        "template_bucket_name",
        "template_key_prefix",
        "required_version",
        "j2_environment",
    },
)

STACK_CONFIG_ATTRIBUTES = ConfigAttributes(
    {},
    {
        "template_path",
        "template",
        "dependencies",
        "dependencies_inheritance",
        "hooks",
        "hooks_inheritance",
        "iam_role",
        "sceptre_role",
        "iam_role_session_duration",
        "sceptre_role_session_duration",
        "notifications",
        "on_failure",
        "parameters",
        "parameters_inheritance",
        "profile",
        "protect",
        "role_arn",
        "cloudformation_service_role",
        "sceptre_user_data",
        "sceptre_user_data_inheritance",
        "stack_name",
        "stack_tags",
        "stack_tags_inheritance",
        "stack_timeout",
    },
)

REQUIRED_KEYS = STACK_GROUP_CONFIG_ATTRIBUTES.required.union(
    STACK_CONFIG_ATTRIBUTES.required
)


class ConfigReader(object):
    """
    Parses YAML configuration files and produces Stack objects.

    Responsible for loading Resolvers and Hook classes and adding them as
    constructors to the PyYAML parser.

    :param context: A SceptreContext.
    :type sceptre.context.SceptreContext:
    """

    def __init__(self, context):
        self.logger = logging.getLogger(__name__)
        self.context = context
        self.full_config_path = self.context.full_config_path()

        # Check is valid sceptre project folder
        self._check_valid_project_path(self.full_config_path)

        # Add Resolver and Hook classes to PyYAML loader
        self._add_yaml_constructors(["sceptre.hooks", "sceptre.resolvers"])
        if not self.context.user_variables:
            self.context.user_variables = {}

        self.templating_vars = {"var": self.context.user_variables}

    @staticmethod
    def _iterate_entry_points(group):
        """
        Helper to determine whether to use pkg_resources or importlib.metadata.
        https://docs.python.org/3/library/importlib.metadata.html
        """
        if sys.version_info < (3, 10):
            from pkg_resources import iter_entry_points

            return iter_entry_points(group)
        else:
            from importlib.metadata import entry_points

            return entry_points(group=group)

    def _add_yaml_constructors(self, entry_point_groups):
        """
        Adds PyYAML constructor functions for all classes found registered at
        the given entry point groups. Classes are registered whereby the node
        tag is the entry point name.

        :param entry_point_groups: Names of entry point groups.
        :type entry_point_groups: list
        """
        self.logger.debug(
            "Adding yaml constructors for the entry point groups {0}".format(
                entry_point_groups
            )
        )

        def constructor_factory(node_class):
            """
            Returns constructor that will initialise objects from a
            given node class.

            :param node_class: Class representing the node.
            :type node_class: class
            :returns: Class initialiser.
            :rtype: func
            """

            # This function signature is required by PyYAML
            def class_constructor(loader, node):
                return node_class(
                    loader.construct_object(self.resolve_node_tag(loader, node))
                )  # pragma: no cover

            return class_constructor

        for group in entry_point_groups:
            for entry_point in self._iterate_entry_points(group):
                # Retrieve name and class from entry point
                node_tag = "!" + entry_point.name
                node_class = entry_point.load()

                # Add constructor to PyYAML loader
                yaml.SafeLoader.add_constructor(
                    node_tag, constructor_factory(node_class)
                )
                self.logger.debug(
                    "Added constructor for %s with node tag %s",
                    str(node_class),
                    node_tag,
                )

    def resolve_node_tag(self, loader, node):
        node = copy.copy(node)
        node.tag = loader.resolve(type(node), node.value, (True, False))
        return node

    def construct_stacks(self) -> Tuple[Set[Stack], Set[Stack]]:
        """
        Traverses the files under the command path.
        For each file encountered, a Stack is constructed
        using the correct config. Dependencies are traversed
        and a final set of Stacks is returned.

        :returns: A set of Stacks.
        """
        stack_map = {}
        command_stacks = set()

        root = self.context.full_command_path()

        if self.context.full_scan:
            root = self.context.full_config_path()

        if path.isfile(root):
            todo = {root}
        else:
            todo = set()
            for directory_name, sub_directories, files in walk(root, followlinks=True):
                for filename in fnmatch.filter(files, "*.yaml"):
                    if filename.startswith("config."):
                        continue

                    todo.add(path.join(directory_name, filename))

        stack_group_configs = {}
        full_todo = todo.copy()
        deps_todo = set()

        while todo:
            abs_path = todo.pop()
            rel_path = path.relpath(abs_path, start=self.context.full_config_path())
            directory, filename = path.split(rel_path)

            if directory in stack_group_configs:
                stack_group_config = stack_group_configs[directory]
            else:
                stack_group_config = stack_group_configs[directory] = self._read(
                    path.join(directory, self.context.config_file)
                )

            stack = self._construct_stack(rel_path, stack_group_config)
            for dep in stack.dependencies:
                full_dep = str(Path(self.context.full_config_path(), dep))
                if not path.exists(full_dep):
                    raise DependencyDoesNotExistError(
                        "{stackname}: Dependency {dep} not found. "
                        "Please make sure that your dependencies stack_outputs "
                        "have their full path from `config` defined.".format(
                            stackname=stack.name, dep=dep
                        )
                    )

                if full_dep not in full_todo and full_dep not in deps_todo:
                    todo.add(full_dep)
                    deps_todo.add(full_dep)

            stack_map[sceptreise_path(rel_path)] = stack

            full_command_path = self.context.full_command_path()
            if abs_path == full_command_path or abs_path.startswith(
                full_command_path.rstrip(path.sep) + path.sep
            ):
                command_stacks.add(stack)

        stacks = self.resolve_stacks(stack_map)

        return stacks, command_stacks

    def resolve_stacks(self, stack_map) -> Set[Stack]:
        """
        Transforms map of Stacks into a set of Stacks, transforms dependencies
        from a list of Strings (stack names) to a list of Stacks.

        :param stack_map: Map of stacks, containing dependencies as list of Strings.
        :type base_config: dict
        :returns: Set of stacks, containing dependencies as list of Stacks.
        :rtype: set
        :raises: sceptre.exceptions.DependencyDoesNotExistError
        """
        stacks = set()
        for stack in stack_map.values():
            if not self.context.ignore_dependencies:
                for i, dep in enumerate(stack.dependencies):
                    try:
                        if not isinstance(dep, Stack):
                            # If the dependency was inherited from a stack group, it might already
                            # have been mapped and so doesn't need to be mapped again.
                            stack.dependencies[i] = stack_map[sceptreise_path(dep)]
                    except KeyError:
                        raise DependencyDoesNotExistError(
                            "{stackname}: Dependency {dep} not found. "
                            "Valid dependency names are: "
                            "{stackkeys}. "
                            "Please make sure that your dependencies stack_outputs "
                            "have their full path from `config` defined.".format(
                                stackname=stack.name,
                                dep=dep,
                                stackkeys=", ".join(stack_map.keys()),
                            )
                        )
                # We deduplicate the dependencies using a set here, since it's possible that a given
                # dependency ends up in the list multiple times.
                stack.dependencies = list(set(stack.dependencies))
            else:
                stack.dependencies = []
            stacks.add(stack)
        return stacks

    def _read(self, rel_path, base_config=None):
        """
        Reads in configuration from one or more YAML files
        within the Sceptre project folder.

        :param rel_path: Relative path to config to read.
        :type rel_path: str
        :param base_config: Base config to provide defaults.
        :type base_config: dict
        :returns: Config read from config files.
        :rtype: dict
        """
        self.logger.debug("Reading in '%s' files...", rel_path)
        directory_path, filename = path.split(rel_path)
        abs_path = path.join(self.full_config_path, rel_path)

        # Adding properties from class
        config = {
            "project_path": self.context.project_path,
            "stack_group_path": directory_path,
        }

        # Adding defaults from base config.
        if base_config:
            config.update(base_config)

        # Check if file exists, but ignore config.yaml as can be inherited.
        if not path.isfile(abs_path) and not filename.endswith(
            self.context.config_file
        ):
            raise ConfigFileNotFoundError(
                'Config file "{0}" not found.'.format(rel_path)
            )

        # Parse and read in the config files.
        this_config = self._recursive_read(directory_path, filename, config)
        # Apply merge strategies with the config that includes base_config values.
        this_config.update(self._get_merge_with_stratgies(config, this_config))
        config.update(this_config)

        self._check_version(config)

        self.logger.debug("Config: %s", config)
        return config

    def _recursive_read(
        self, directory_path: str, filename: str, stack_group_config: dict
    ) -> dict:
        """
        Traverses the directory_path, from top to bottom, reading in all
        relevant config files. If config attributes are encountered further
        down the StackGroup they are merged with the parent as defined in the
        `CONFIG_MERGE_STRATEGIES` dict.

        :param directory_path: Relative directory path to config to read.
        :param filename: File name for the config to read.
        :param stack_group_config: The loaded config file for the StackGroup
        :returns: Representation of inherited config.
        """

        parent_directory = path.split(directory_path)[0]

        # Base condition for recursion
        config = {}

        if directory_path:
            config = self._recursive_read(
                parent_directory, filename, stack_group_config
            )

        # Combine the stack_group_config with the nested config dict
        config_group = stack_group_config.copy()
        config_group.update(config)

        # Read config file and overwrite inherited properties
        child_config = self._render(directory_path, filename, config_group) or {}
        child_config.update(self._get_merge_with_stratgies(config, child_config))
        config.update(child_config)
        return config

    def _get_merge_with_stratgies(self, left: dict, right: dict) -> dict:
        """
        Returns a new dict with only the merge values of the two inputs, using the
        merge strategies defined for each key.
        """
        merge = {}

        # Then apply the merge strategies to each item
        for config_key, default_strategy in CONFIG_MERGE_STRATEGIES.items():
            strategy = default_strategy
            override_key = f"{config_key}_inheritance"
            if override_key in CONFIG_MERGE_STRATEGIES:
                name = CONFIG_MERGE_STRATEGIES[override_key](
                    left.get(override_key), right.get(override_key)
                )
                if not name:
                    pass
                elif name not in CONFIG_MERGE_STRATEGY_OVERRIDES[config_key]:
                    raise SceptreException(
                        f"{name} is not a valid inheritance strategy for {config_key}"
                    )
                else:
                    strategy = CONFIG_MERGE_STRATEGY_OVERRIDES[config_key][name]

            value = strategy(left.get(config_key), right.get(config_key))
            if value:
                merge[config_key] = value

        return merge

    def _render(self, directory_path, basename, stack_group_config):
        """
        Reads a configuration file, loads the config file as a template
        and returns config loaded from the file.

        :param directory_path: Relative directory path to config to read.
        :type directory_path: str
        :param basename: The filename of the config file
        :type basename: str
        :param stack_group_config: The loaded config file for the StackGroup
        :type stack_group_config: dict
        :returns: rendered template of config file.
        :rtype: dict
        """
        config = {}
        abs_directory_path = path.join(self.full_config_path, directory_path)

        if not path.isfile(path.join(abs_directory_path, basename)):
            return

        default_j2_environment_config = {
            "autoescape": select_autoescape(
                disabled_extensions=("yaml",),
                default=True,
            ),
            "loader": FileSystemLoader(abs_directory_path),
            "undefined": StrictUndefined,
        }
        j2_environment_config = strategies.dict_merge(
            default_j2_environment_config,
            stack_group_config.get("j2_environment", {}),
        )
        j2_environment = Environment(**j2_environment_config)

        try:
            template = j2_environment.get_template(basename)
        except Exception as err:
            raise SceptreException(
                f"{Path(directory_path, basename).as_posix()} - {err}"
            ) from err

        # Reset the template cache to avoid leakage between StackGroups (#937)
        template_vars = {"var": self.templating_vars["var"]}
        if "stack_group_config" in self.templating_vars:
            template_vars["stack_group_config"] = self.templating_vars[
                "stack_group_config"
            ]
        self.templating_vars = template_vars

        self.templating_vars.update(stack_group_config)

        try:
            rendered_template = template.render(
                self.templating_vars,
                command_path=self.context.command_path.split(path.sep),
                environment_variable=environ,
            )
        except Exception as err:
            message = f"{Path(directory_path, basename).as_posix()} - {err}"

            if logging_level() == logging.DEBUG:
                debug_file_path = write_debug_file(
                    json.dumps(self.templating_vars, indent=4), prefix="vars_"
                )
                message += f"\nTemplating vars saved to: {debug_file_path}"

            raise SceptreException(message) from err

        try:
            config = yaml.safe_load(rendered_template)
        except Exception as err:
            message = f"Error parsing {abs_directory_path}{basename}:\n{err}"

            if logging_level() == logging.DEBUG:
                debug_file_path = write_debug_file(
                    rendered_template, prefix="rendered_"
                )
                message += f"\nRendered template saved to: {debug_file_path}"

            raise ValueError(message)

        return config

    @staticmethod
    def _check_valid_project_path(config_path):
        """
        Raises an InvalidSceptreDirectoryError if ``path`` is not a directory.

        :param path: A config directory path.
        :type path: str
        :raises: sceptre.exceptions.InvalidSceptreDirectoryError
        """
        if not path.isdir(config_path):
            raise InvalidSceptreDirectoryError(
                "Check '{0}' exists.".format(config_path)
            )

    def _check_version(self, config):
        """
        Raises a VersionIncompatibleException when the current Sceptre version
        does not comply with the configured version requirement.

        :raises: sceptre.exceptions.VersionIncompatibleException
        """
        sceptre_version = __version__
        if "required_version" in config:
            required_version = config["required_version"]
            if Version(sceptre_version) not in SpecifierSet(required_version, True):
                raise VersionIncompatibleError(
                    "Current sceptre version ({0}) does not meet version "
                    "requirements: {1}".format(sceptre_version, required_version)
                )

    @staticmethod
    def _collect_s3_details(stack_name, config):
        """
        Collects and constructs details for where to store the Template in S3.

        :param stack_name: Stack name.
        :type stack_name: str
        :param config: Config with details.
        :type config: dict
        :returns: S3 details.
        :rtype: dict
        """
        s3_details = None
        # If the config explicitly sets the template_bucket_name to None, we don't want to enter
        # this conditional block.
        if config.get("template_bucket_name") is not None:
            template_key = "/".join(
                [
                    sceptreise_path(stack_name),
                    "{time_stamp}.json".format(
                        time_stamp=datetime.datetime.utcnow().strftime(
                            "%Y-%m-%d-%H-%M-%S-%fZ"
                        )
                    ),
                ]
            )

            if "template_key_prefix" in config:
                prefix = config["template_key_prefix"]
                template_key = "/".join([prefix.strip("/"), template_key])

            s3_details = {
                "bucket_name": config["template_bucket_name"],
                "bucket_key": template_key,
            }
        return s3_details

    def _construct_stack(self, rel_path, stack_group_config=None):
        """
        Constructs an individual Stack object from a config path and a
        base config.

        :param rel_path: A relative config file path.
        :type rel_path: str
        :param stack_group_config: The Stack group config to use as defaults.
        :type stack_group_config: dict
        :returns: Stack object
        :rtype: sceptre.stack.Stack
        """

        directory, filename = path.split(rel_path)
        if filename == self.context.config_file:
            pass

        self.templating_vars["stack_group_config"] = stack_group_config
        parsed_stack_group_config = self._parsed_stack_group_config(stack_group_config)
        config = self._read(rel_path, stack_group_config)
        stack_name = path.splitext(rel_path)[0]

        # Check for missing mandatory attributes
        for required_key in REQUIRED_KEYS:
            if required_key not in config:
                raise InvalidConfigFileError(
                    "Required attribute '{0}' not found in configuration of '{1}'.".format(
                        required_key, stack_name
                    )
                )

        s3_details = self._collect_s3_details(stack_name, config)
        # If disable/enable rollback was specified on the command line, use that. Otherwise,
        # fall back to the stack config.
        disable_rollback = self.context.command_params.get("disable_rollback")
        if disable_rollback is None:
            disable_rollback = config.get("disable_rollback", False)

        stack = Stack(
            name=stack_name,
            project_code=config["project_code"],
            template_path=config.get("template_path"),
            template_handler_config=config.get("template"),
            region=config["region"],
            template_bucket_name=config.get("template_bucket_name"),
            template_key_prefix=config.get("template_key_prefix"),
            required_version=config.get("required_version"),
            sceptre_role=config.get("sceptre_role"),
            iam_role=config.get("iam_role"),
            sceptre_role_session_duration=config.get("sceptre_role_session_duration"),
            iam_role_session_duration=config.get("iam_role_session_duration"),
            profile=config.get("profile"),
            parameters=config.get("parameters", {}),
            sceptre_user_data=config.get("sceptre_user_data", {}),
            hooks=config.get("hooks", {}),
            s3_details=s3_details,
            dependencies=config.get("dependencies", []),
            role_arn=config.get("role_arn"),
            cloudformation_service_role=config.get("cloudformation_service_role"),
            protected=config.get("protect", False),
            tags=config.get("stack_tags", {}),
            external_name=config.get("stack_name"),
            notifications=config.get("notifications"),
            on_failure=config.get("on_failure"),
            disable_rollback=disable_rollback,
            stack_timeout=config.get("stack_timeout", 0),
            ignore=config.get("ignore", False),
            obsolete=config.get("obsolete", False),
            stack_group_config=parsed_stack_group_config,
            config=config,
        )

        del self.templating_vars["stack_group_config"]
        return stack

    def _parsed_stack_group_config(self, stack_group_config):
        """
        Remove all config items that are supported by Sceptre and
        remove the `project_path` and `stack_group_path` added by `read()`.
        Return a dictionary that has only user-specified config items.
        """
        parsed_config = {
            key: stack_group_config[key]
            for key in set(stack_group_config) - set(CONFIG_MERGE_STRATEGIES)
        }
        parsed_config.pop("stack_group_path")
        return parsed_config



import logging
import click

from pathlib import Path

from sceptre.context import SceptreContext
from sceptre.cli.helpers import catch_exceptions, write
from sceptre.plan.plan import SceptrePlan
from sceptre.helpers import null_context
from sceptre.resolvers.placeholders import use_resolver_placeholders_on_error

logger = logging.getLogger(__name__)


@click.group(name="dump")
def dump_group():
    """
    Commands for dumping attributes of stacks.
    """
    pass


@dump_group.command(name="config")
@click.argument("path")
@click.option(
    "--to-file", is_flag=True, help="If True, also dump the template to a local file."
)
@click.pass_context
@catch_exceptions
def dump_config(ctx, to_file, path):
    """
    Dump the rendered (post-Jinja) Stack Configs.
    \f

    :param path: Path to execute the command on or path to stack group
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        output_format=ctx.obj.get("output_format"),
        options=ctx.obj.get("options"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )
    plan = SceptrePlan(context)
    responses = plan.dump_config()

    output_format = "json" if context.output_format == "json" else "yaml"

    for stack, config in responses.items():
        stack_name = stack.external_name

        if to_file:
            file_path = Path(".dump") / stack_name / f"config.{output_format}"
            logger.info(f"{stack_name} dumping to {file_path}")
            write(config, output_format, no_colour=True, file_path=file_path)
            logger.info(f"{stack_name} dump to {file_path} complete.")

        else:
            write(config, output_format, no_colour=True)


@dump_group.command(name="template")
@click.argument("path")
@click.option(
    "-n",
    "--no-placeholders",
    is_flag=True,
    help="If True, no placeholder values will be supplied for resolvers that cannot be resolved.",
)
@click.option(
    "--to-file", is_flag=True, help="If True, also dump the template to a local file."
)
@click.pass_context
@catch_exceptions
def dump_template(ctx, to_file, no_placeholders, path):
    """
    Prints the template used for stack in PATH.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    context = SceptreContext(
        command_path=path,
        command_params=ctx.params,
        project_path=ctx.obj.get("project_path"),
        user_variables=ctx.obj.get("user_variables"),
        options=ctx.obj.get("options"),
        output_format=ctx.obj.get("output_format"),
        ignore_dependencies=ctx.obj.get("ignore_dependencies"),
    )
    plan = SceptrePlan(context)

    execution_context = (
        null_context() if no_placeholders else use_resolver_placeholders_on_error()
    )
    with execution_context:
        responses = plan.dump_template()

    output_format = "json" if context.output_format == "json" else "yaml"

    for stack, template in responses.items():
        stack_name = stack.external_name

        if to_file:
            file_path = Path(".dump") / stack_name / f"template.{output_format}"
            logger.info(f"{stack_name} dumping template to {file_path}")
            write(template, output_format, no_colour=True, file_path=file_path)
            logger.info(f"{stack_name} dump to {file_path} complete.")

        else:
            write(template, output_format, no_colour=True)


@dump_group.command(name="all")
@click.argument("path")
@click.option(
    "-n",
    "--no-placeholders",
    is_flag=True,
    help="If True, no placeholder values will be supplied for resolvers that cannot be resolved.",
)
@click.option(
    "--to-file", is_flag=True, help="If True, also dump the template to a local file."
)
@click.pass_context
@catch_exceptions
def dump_all(ctx, to_file, no_placeholders, path):
    """
    Dumps both the rendered (post-Jinja) Stack Configs and the template used for stack in PATH.
    \f

    :param path: Path to execute the command on.
    :type path: str
    """
    ctx.invoke(dump_config, to_file=to_file, path=path)
    ctx.invoke(
        dump_template, to_file=to_file, no_placeholders=no_placeholders, path=path
    )
